/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines classes @link ders::mem_buf mem_buf @endlink, @link ders::buf_cache
 * buf_cache @endlink.
 */

#ifndef __DERS__BUF_CACHE_HPP__
#define __DERS__BUF_CACHE_HPP__

#include <ders/config.hpp>
#include <vector>

namespace ders {  // ::ders

/**
 * Helper class for buf_cache to ensure exception safety.
 */
class mem_buf {
      void* buf;

      mem_buf(const mem_buf&);
      mem_buf& operator=(const mem_buf&);

 public:
      mem_buf(void* bf) : buf(bf) {}
      ~mem_buf() { operator delete(buf); }

      void* get() const { return buf; }
      void release() { buf=0; }
};

/**
 * Keeps large buffers of requested size which are too expensive to allocate on
 * demand.
 */
class buf_cache {
      std::vector<void*> bufs;
      int bsz;

      buf_cache(const buf_cache&);
      buf_cache& operator=(const buf_cache&);

 public:
      buf_cache(int bsize) : bsz(bsize) {}
      ~buf_cache();

      void* take_buf();
      void return_buf(mem_buf& mb);
      int buf_size() const { return bsz; }
};

inline buf_cache::~buf_cache()
{
 for (int i=0, end=bufs.size(); i<end; i++) operator delete(bufs[i]);
}

inline void* buf_cache::take_buf()
{
 if (!bufs.size()) return operator new(bsz);

 void* ret=bufs.back();
 bufs.pop_back();
 return ret;
}

inline void buf_cache::return_buf(mem_buf& mb)
{
 bufs.push_back(mb.get());
 mb.release();
}

}  // namespace ::ders

#endif  // __DERS__BUF_CACHE_HPP__

