/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines interfaces @link ders::readable readable @endlink, @link
 * ders::writable writable @endlink and
 * classes @link ders::buf_reader buf_reader @endlink, @link ders::buf_writer
 * buf_writer @endlink.
 */

#ifndef __DERS__BUF_RDWR_HPP__
#define __DERS__BUF_RDWR_HPP__

#include <ders/config.hpp>
#include <ders/text.hpp>

namespace ders {  // ::ders

/**
 * Interface for %readable classes.
 */
class readable : virtual public destroyable {
 public:
      virtual int read(void* buf, int n)=0;
};

/**
 * Interface for %writable classes.
 */
class writable : virtual public destroyable {
 public:
      virtual int write(const void* buf, int n)=0;
};

/**
 * Provides buffered read functions for classes that implement readable
 * interface.
 */
class buf_reader {
      mem_pool& mp;
      readable& rd;
      unsigned char* buf;
      int bsz;
      int it, lim;

      int read_impl();

      buf_reader(const buf_reader&);
      buf_reader& operator=(const buf_reader&);

 public:
      buf_reader(mem_pool& m, readable& rdbl, int bufsz=4*1024);
      ~buf_reader() { mp.deallocate(buf, bsz); }

      int read(void* b, int n);
      int read(sh_text& buf, int n);
      int read() { return (it<lim) ? buf[it++] : read_impl(); }
      bool read_line(sh_text& line, bool noendl=true);

      void reset() { it=lim=0; }

      const char* begin() const { return reinterpret_cast<char*>(buf)+it; }
      const char* end() const { return reinterpret_cast<char*>(buf)+lim; }
};

/**
 * Provides buffered write functions for classes that implement writable
 * interface.
 */
class buf_writer {
      mem_pool& mp;
      writable& wr;
      unsigned char* buf;
      int bsz;
      int it;

      int write_impl(char ch);

      buf_writer(const buf_writer&);
      buf_writer& operator=(const buf_writer&);

 public:
      buf_writer(mem_pool& m, writable& wrbl, int bufsz=4*1024);
      ~buf_writer();

      int write(const void* b, int n);
      int write(const ch_rng& r) { return write(r.beg, r.size()); }
      int write(char ch) { return (it<bsz) ? buf[it++]=ch : write_impl(ch); }
      bool flush();

      void reset() { it=0; }

      const char* begin() const { return reinterpret_cast<char*>(buf); }
      const char* end() const { return reinterpret_cast<char*>(buf)+it; }
};

}  // namespace ::ders

#endif  // __DERS__BUF_RDWR_HPP__

