/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines interfaces @link ders::data_in data_in @endlink, @link ders::data_out
 * data_out @endlink, @link ders::data_io data_io @endlink and
 * function new_data_io().
 */

#ifndef __DERS__DATA_IO_HPP__
#define __DERS__DATA_IO_HPP__

#include <ders/config.hpp>
#include <ders/cycl_buf.hpp>
#include <ders/text.hpp>

namespace ders {  // ::ders

class object_io;

/**
 * Interface for classes that can be read from.
 */
class data_in : virtual public destroyable {
 public:
      virtual bool read_bool(bool& val)=0;

      virtual bool read_char(char& val)=0;
      virtual bool read_schar(signed char& val)=0;
      virtual bool read_uchar(unsigned char& val)=0;
      virtual bool read_wchar(wchar_t& val)=0;

      virtual bool read_short(short& val)=0;
      virtual bool read_ushort(unsigned short& val)=0;

      virtual bool read_int(int& val)=0;
      virtual bool read_uint(unsigned int& val)=0;

      virtual bool read_long(long& val)=0;
      virtual bool read_ulong(unsigned long& val)=0;

      virtual bool read_float(float& val)=0;
      virtual bool read_double(double& val)=0;
      virtual bool read_ldouble(long double& val)=0;

      virtual bool read_ptr(void*& val)=0;
      template<class T> bool read_ptr(T*& val);

      virtual bool read_text(sh_text& val)=0;
};

/**
 * Interface for classes that can be written to.
 */
class data_out : virtual public destroyable {
 public:
      virtual void write_bool(bool val)=0;

      virtual void write_char(char val)=0;
      virtual void write_schar(signed char val)=0;
      virtual void write_uchar(unsigned char val)=0;
      virtual void write_wchar(wchar_t val)=0;

      virtual void write_short(short val)=0;
      virtual void write_ushort(unsigned short val)=0;

      virtual void write_int(int val)=0;
      virtual void write_uint(unsigned int val)=0;

      virtual void write_long(long val)=0;
      virtual void write_ulong(unsigned long val)=0;

      virtual void write_float(float val)=0;
      virtual void write_double(double val)=0;
      virtual void write_ldouble(long double val)=0;

      virtual void write_ptr(void* val)=0;

      virtual void write_text(const ch_rng& val)=0;
};

/**
 * Interface for classes that support I/O operations defined by both data_in and
 * data_out interfaces.
 */
class data_io : public data_in, public data_out {
};

typedef sh_ptr<data_io> sh_data_io;

/**
 * Creates new data_io object.
 * @param cb cycl_buf to use.
 * @return created object.
 */
sh_data_io new_data_io(sh_cycl_buf cb);

template<class T>
inline bool data_in::read_ptr(T*& val)
{
 void* ptr;
 bool ret=read_ptr(ptr);
 val=static_cast<T*>(ptr);

 return ret;
}

}  // namespace ::ders

#endif  // __DERS__DATA_IO_HPP__

