/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines interface @link ders::destroyable destroyable @endlink and
 * function @link ders::destroy() destroy() @endlink.
 */

#ifndef __DERS__DESTROY_HPP__
#define __DERS__DESTROY_HPP__

#include <ders/config.hpp>
#include <typeinfo>
#include <assert.h>
#include <ders/hard_asrt.hpp>
#include <ders/mem_pool.hpp>

namespace ders {  // ::ders

/**
 * Interface for %destroyable classes. All interfaces must be virtually derived
 * from this base class.
 */
class destroyable {
      destroyable& operator=(const destroyable&);
      void* operator new(size_t size);

 protected:
      template<class T>
      static void destroy_this(T* ptr, mem_pool& mp2);

      void operator delete(void* ptr, size_t size) { hard_assert(false); }

 public:
      virtual void destroy(mem_pool& mp2)=0;
      virtual ~destroyable() {}
};

template<class T>
inline void destroyable::destroy_this(T* ptr, mem_pool& mp2)
{
 assert(typeid(*ptr)==typeid(T));

 ptr->~T();
 mp2.deallocate(ptr, sizeof(T));
}

namespace detail {  // ::ders::detail

template<class T>
inline void destroy(void* p, mem_pool& mp)
{
 T* ptr=static_cast<T*>(p);

 ptr->~T();
 mp.deallocate(ptr, sizeof(T));
}

template<class T>
inline void destroy(destroyable* ptr, mem_pool& mp)
{ ptr->destroy(mp); }

}  // namespace ::ders::detail

/**
 * Placement delete for mem_pool allocated objects. Destroys passed object
 * depending on whether it implements destroyable interface or not.
 * @param ptr points to object to destroy. Can be NULL.
 * @param mp mem_pool to use.
 */
template<class T>
inline void destroy(T* ptr, mem_pool& mp)
{ if (ptr) detail::destroy<T>(ptr, mp); }

}  // namespace ::ders

#endif  // __DERS__DESTROY_HPP__

