/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines enum @link ders::error error @endlink and
 * functions error_message(), convert_errno() and
 * exception class @link ders::ErrorException ErrorException @endlink.
 */

#ifndef __DERS__ERROR_HPP__
#define __DERS__ERROR_HPP__

#include <ders/config.hpp>
#include <ders/exception.hpp>

namespace ders {  // ::ders

/**
 * Set of error codes that don't depend on underlining implementation.
 */
enum error {
  ezero=           0,  // No error
  e2big=           1,  // Argument list too long
  eacces=          2,  // Permission denied
  eaddrinuse=      3,  // Address in use
  eaddrnotavail=   4,  // Address not available
  eafnosupport=    5,  // Address family not supported
  eagain=          6,  // Resource unavailable, try again
  ealready=        7,  // Connection already in progress
  ebadf=           8,  // Bad file descriptor
  ebadmsg=         9,  // Bad message
  ebusy=          10,  // Device or resource busy
  ecanceled=      11,  // Operation canceled
  echild=         12,  // No child processes
  econnaborted=   13,  // Connection aborted
  econnrefused=   14,  // Connection refused
  econnreset=     15,  // Connection reset
  edeadlk=        16,  // Resource deadlock would occur
  edestaddrreq=   17,  // Destination address required
  edom=           18,  // Mathematics argument out of domain of function
  edquot=         19,  // Reserved (EDQUOT)
  eexist=         20,  // File exists
  efault=         21,  // Bad address
  efbig=          22,  // File too large
  ehostunreach=   23,  // Host is unreachable
  eidrm=          24,  // Identifier removed
  eilseq=         25,  // Illegal byte sequence
  einprogress=    26,  // Operation in progress
  eintr=          27,  // Interrupted function
  einval=         28,  // Invalid argument
  eio=            29,  // I/O error
  eisconn=        30,  // Socket is connected
  eisdir=         31,  // Is a directory
  eloop=          32,  // Too many levels of symbolic links
  emfile=         33,  // Too many open files
  emlink=         34,  // Too many links
  emsgsize=       35,  // Message too large
  emultihop=      36,  // Reserved (EMULTIHOP)
  enametoolong=   37,  // Filename too long
  enetdown=       38,  // Network is down
  enetreset=      39,  // Connection aborted by network
  enetunreach=    40,  // Network unreachable
  enfile=         41,  // Too many files open in system
  enobufs=        42,  // No buffer space available
  enodata=        43,  // No message is available on the STREAM head read queue
  enodev=         44,  // No such device
  enoent=         45,  // No such file or directory
  enoexec=        46,  // Executable file format error
  enolck=         47,  // No locks available
  enolink=        48,  // Reserved (ENOLINK)
  enomem=         49,  // Not enough space
  enomsg=         50,  // No message of the desired type
  enoprotoopt=    51,  // Protocol not available
  enospc=         52,  // No space left on device
  enosr=          53,  // No STREAM resources
  enostr=         54,  // Not a STREAM
  enosys=         55,  // Function not supported
  enotconn=       56,  // The socket is not connected
  enotdir=        57,  // Not a directory
  enotempty=      58,  // Directory not empty
  enotsock=       59,  // Not a socket
  enotsup=        60,  // Not supported
  enotty=         61,  // Inappropriate I/O control operation
  enxio=          62,  // No such device or address
  eopnotsupp=     63,  // Operation not supported on socket
  eoverflow=      64,  // Value too large to be stored in data type
  eperm=          65,  // Operation not permitted
  epipe=          66,  // Broken pipe
  eproto=         67,  // Protocol error
  eprotonosupport=68,  // Protocol not supported
  eprototype=     69,  // Protocol wrong type for socket
  erange=         70,  // Result too large
  erofs=          71,  // Read-only file system
  espipe=         72,  // Invalid seek
  esrch=          73,  // No such process
  estale=         74,  // Reserved (ESTALE)
  etime=          75,  // Stream ioctl() timeout
  etimedout=      76,  // Connection timed out
  etxtbsy=        77,  // Text file busy
  eunkn=          78,  // Unknown error
  ewouldblock=    79,  // Operation would block
  exdev=          80   // Cross-device link
};

/**
 * Returns error message for passed error.
 * @param mp mem_pool to use.
 * @param err error code.
 * @return error message.
 */
sh_text error_message(mem_pool& mp, error err);

/**
 * Converts implementation dependand errno value to error code.
 * @param val errno value.
 * @return error code.
 */
error convert_errno(int val);

shException newErrorException(mem_pool& mp, const FileLine& loc, const ch_rng&
  msg, error err);
shException newErrorException(const FileLine& loc, const ch_rng& msg, error err,
  shException nest);

/**
 * %Exception class with embed errorCode.
 */
class ErrorException : public Exception {
      friend shException newErrorException(const FileLine& loc, const ch_rng&
        msg, error err, shException nest);

      ErrorException(const ErrorException&);
      ErrorException& operator=(const ErrorException&);

 protected:
      ErrorException(const FileLine& loc, const ch_rng& msg, error err,
        shException nest);

 public:
      const error errorCode;

      virtual sh_text getClassName() const;
      virtual sh_text toText() const;

      virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
};

}  // namespace ::ders

#endif  // __DERS__ERROR_HPP__

