/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines struct @link ders::FileLine FileLine @endlink and
 * macro #_FLINE_ and
 * functions toTextAll(), recatchException() and
 * exception classes @link ders::Exception Exception @endlink, @link
 * ders::MsgException MsgException @endlink, @link ders::ExitMsgException
 * ExitMsgException @endlink, @link ders::ExternException ExternException
 * @endlink, @link ders::StdExternException StdExternException @endlink, @link
 * ders::UnknExternException UnknExternException @endlink.
 */

#ifndef __DERS__EXCEPTION_HPP__
#define __DERS__EXCEPTION_HPP__

#include <ders/config.hpp>
#include <ders/text.hpp>

/**
 * Helper macro for automatic FileLine object creation.
 */
#define _FLINE_ ders::FileLine(__FILE__, __LINE__)

namespace ders {  // ::ders

class Exception;
typedef sh_ptr<Exception> shException;

/**
 * Encapsulates %file name and line number information.
 */
struct FileLine {
       const char* file;
       int line;

       FileLine(const char* f, int l) : file(f), line(l) {}
};

/**
 * Returns concatenated error message of all nested exceptions.
 * @param she exception object.
 * @return error message.
 */
sh_text toTextAll(const shException& she);

/**
 * Converts current exception to shException object.
 * @param mp mem_pool to use.
 * @param location location to use for non-shException exceptions.
 * @return current or newly created exception object.
 */
shException recatchException(mem_pool& mp, const FileLine& location);

/**
 * Base exception class with embed location, message and nested exception.
 */
class Exception : public destroyable {
      friend shException newException(const FileLine& loc, const ch_rng& msg,
        shException nest);

      Exception(const Exception&);
      Exception& operator=(const Exception&);

 protected:
      Exception(const FileLine& loc, const ch_rng& msg, shException nest);

 public:
      const FileLine location;
      const sh_text message;
      const shException nested;

      virtual sh_text getClassName() const;
      virtual sh_text toText() const;

      template<class E> E* is();
      template<class E> const E* is() const;

      virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
};

shException newException(mem_pool& mp, const FileLine& loc, const ch_rng& msg);
shException newException(const FileLine& loc, const ch_rng& msg, shException
  nest);

/**
 * Abstract exception class for message exceptions that must pass exception
 * handlers untouched.
 */
class MsgException : public Exception {
      MsgException(const MsgException&);
      MsgException& operator=(const MsgException&);

 protected:
      MsgException(const FileLine& loc, const ch_rng& msg, shException nest);

 public:
      virtual sh_text getClassName() const=0;
};

/**
 * %MsgException class with embed exitCode used to exit application with.
 */
class ExitMsgException : public MsgException {
      friend shException newExitMsgException(const FileLine& loc, const ch_rng&
        msg, int ecode, shException nest);

      ExitMsgException(const ExitMsgException&);
      ExitMsgException& operator=(const ExitMsgException&);

 protected:
      ExitMsgException(const FileLine& loc, const ch_rng& msg, int ecode,
        shException nest);

 public:
      const int exitCode;

      virtual sh_text getClassName() const;
      virtual sh_text toText() const;

      virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
};

shException newExitMsgException(mem_pool& mp, const FileLine& loc, const ch_rng&
  msg, int ecode);
shException newExitMsgException(const FileLine& loc, const ch_rng& msg, int
  ecode, shException nest);

/**
 * Abstract exception class for external exceptions with embed typeName.
 */
class ExternException : public Exception {
      ExternException(const ExternException&);
      ExternException& operator=(const ExternException&);

 protected:
      ExternException(const FileLine& loc, const ch_rng& msg, const ch_rng&
        tname, shException nest);

 public:
      const sh_text typeName;

      virtual sh_text getClassName() const=0;
      virtual sh_text toText() const;
};

/**
 * %ExternException class for std::exception.
 */
class StdExternException : public ExternException {
      friend shException newStdExternException(const FileLine& loc, const
        ch_rng& msg, const ch_rng& tname, shException nest);

      StdExternException(const StdExternException&);
      StdExternException& operator=(const StdExternException&);

 protected:
      StdExternException(const FileLine& loc, const ch_rng& msg, const ch_rng&
        tname, shException nest);

 public:
      virtual sh_text getClassName() const;

      virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
};

shException newStdExternException(mem_pool& mp, const FileLine& loc, const
  ch_rng& msg, const ch_rng& tname);
shException newStdExternException(const FileLine& loc, const ch_rng& msg, const
  ch_rng& tname, shException nest);

/**
 * %ExternException class for unknown exception.
 */
class UnknExternException : public ExternException {
      friend shException newUnknExternException(const FileLine& loc, const
        ch_rng& msg, shException nest);

      UnknExternException(const UnknExternException&);
      UnknExternException& operator=(const UnknExternException&);

 protected:
      UnknExternException(const FileLine& loc, const ch_rng& msg, shException
        nest);

 public:
      virtual sh_text getClassName() const;

      virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
};

shException newUnknExternException(mem_pool& mp, const FileLine& loc, const
  ch_rng& msg);
shException newUnknExternException(const FileLine& loc, const ch_rng& msg,
  shException nest);

template<class E> inline E* Exception::is()
{ return dynamic_cast<E*>(this); }

template<class E> inline const E* Exception::is() const
{ return dynamic_cast<const E*>(this); }

}  // namespace ::ders

#endif  // __DERS__EXCEPTION_HPP__

