/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines struct @link ders::path_elmt path_elmt @endlink and
 * character @link ders::pathSepr pathSepr @endlink and
 * set of functions for path processing.
 */

#ifndef __DERS__PATH_HPP__
#define __DERS__PATH_HPP__

#include <ders/config.hpp>
#include <vector>
#include <ders/text.hpp>

namespace ders {  // ::ders

/**
 * Path element for path processing functions.
 */
struct path_elmt {
       /** element type */
       enum type {
            drive,  // "A:"            : drive [root] [name ...]
            name,   // "name"          : name [name ...]
            root,   // "\"             : root [name ...]
            share   // "\\server\share": share [name ...]
       };

       type tp;
       ch_rng nm;

       path_elmt(type t, const ch_rng& n) : tp(t), nm(n) {}
};

/**
 * Implementation-dependant path separator.
 */
extern const char pathSepr;

/**
 * Splits path into path elements.
 * @param mp mem_pool to use.
 * @param path path to parse.
 * @param[out] ret parsed path elements.
 */
void parse_path(mem_pool& mp, const ch_rng& path, std::vector<path_elmt>& ret);

/**
 * Creates path from path elements.
 * @param mp mem_pool to use.
 * @param beg points to first path element.
 * @param end points to past-the-end path element.
 * @param spr path separator to use.
 * @return created path.
 */
sh_text make_path(mem_pool& mp, const path_elmt* beg, const path_elmt* end, char
  spr=pathSepr);

/**
 * Normalizes path via applying parse_path() and make_path() functions.
 * @param mp mem_pool to use.
 * @param path path to normalize.
 * @return normalized path.
 */
sh_text normalize_path(mem_pool& mp, const ch_rng& path);

/**
 * Returns path part of the path.
 * @param mp mem_pool to use.
 * @param path path to parse.
 * @return path part.
 */
sh_text get_path(mem_pool& mp, const ch_rng& path);

/**
 * Returns name part of the path.
 * @param mp mem_pool to use.
 * @param path path to parse.
 * @return name part.
 */
sh_text get_name(mem_pool& mp, const ch_rng& path);

}  // namespace ::ders

#endif  // __DERS__PATH_HPP__

