/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines class @link ders::sh_ptr sh_ptr @endlink.
 */

#ifndef __DERS__SH_PTR_HPP__
#define __DERS__SH_PTR_HPP__

#include <ders/config.hpp>
#include <ders/destroy.hpp>

namespace ders {  // ::ders

namespace detail {  // ::ders::detail

template<class T>
struct sh_ptr_core {
       mem_pool& mp;
       T* ptr;
       int refs;

       sh_ptr_core(mem_pool& m, T* p) : mp(m), ptr(p), refs(1) {}
       ~sh_ptr_core() { ders::destroy(ptr, mp); }
};

}  // namespace ::ders::detail

/**
 * Shared smart pointer that uses destroy() function to destruct passed objects.
 */
template<class T>
class sh_ptr {
      typedef detail::sh_ptr_core<T> core;
      core* cr;

 public:
      sh_ptr(mem_pool& mp, T* ptr);

      sh_ptr(const sh_ptr& shp)
      {
       cr=shp.cr;
       cr->refs++;
      }

      ~sh_ptr();

      sh_ptr& operator=(const sh_ptr& shp);

      mem_pool& pool() const { return cr->mp; }

      void swap(sh_ptr& shp)
      {
       core* tmp=cr;
       cr=shp.cr;
       shp.cr=tmp;
      }

      T& operator*() { return *cr->ptr; }
      const T& operator*() const { return *cr->ptr; }

      T* operator->() { return cr->ptr; }
      const T* operator->() const { return cr->ptr; }

      T* get() { return cr->ptr; }
      const T* get() const { return cr->ptr; }

      void set(T* ptr) { cr->ptr=ptr; }

      int refs() const { return cr->refs; }
};

template<class T>
sh_ptr<T>::sh_ptr(mem_pool& mp, T* ptr)
{
 try { cr=new(mp.allocate(sizeof(core))) core(mp, ptr); }
 catch (...) {
       destroy(ptr, mp);
       throw;
 }
}

template<class T>
sh_ptr<T>::~sh_ptr()
{
 if (--cr->refs==0) destroy(cr, cr->mp);
}

template<class T>
sh_ptr<T>& sh_ptr<T>::operator=(const sh_ptr& shp)
{
 shp.cr->refs++;
 if (--cr->refs==0) destroy(cr, cr->mp);
 cr=shp.cr;

 return *this;
}

}  // namespace ::ders

#endif  // __DERS__SH_PTR_HPP__

