/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines struct @link ders::ch_rng ch_rng @endlink and
 * class @link ders::text text @endlink and
 * set of functions for ch_rng comparison.
 */

#ifndef __DERS__TEXT_HPP__
#define __DERS__TEXT_HPP__

#include <ders/config.hpp>
#include <ders/sh_ptr.hpp>

namespace ders {  // ::ders

struct ch_rng;

class text;
typedef sh_ptr<text> sh_text;

sh_text nt(mem_pool& mp);
sh_text nt(mem_pool& mp, const char* str);
sh_text nt(mem_pool& mp, const char* b, const char* e);
sh_text nt(mem_pool& mp, const ch_rng& r);
sh_text nt(mem_pool& mp, const text& t);
sh_text nt(mem_pool& mp, const sh_text& sht);

bool operator==(const ch_rng& r1, const ch_rng& r2);
bool operator!=(const ch_rng& r1, const ch_rng& r2);
bool operator< (const ch_rng& r1, const ch_rng& r2);
bool operator<=(const ch_rng& r1, const ch_rng& r2);
bool operator> (const ch_rng& r1, const ch_rng& r2);
bool operator>=(const ch_rng& r1, const ch_rng& r2);

/**
 * Represents character range.
 */
struct ch_rng {
       const char* beg;
       const char* end;

       ch_rng(const char* str);
       ch_rng(const char* b, const char* e);
       ch_rng(const text& t);
       ch_rng(const sh_text& sht);

       int size() const { return end-beg; }
};

/**
 * Represents %text or string as sequence of arbitrary characters.
 */
class text {
      mem_pool& mp;
      char* buf;
      int resd;
      int used;

      void init(const ch_rng& r);

 public:
      explicit text(mem_pool& m) : mp(m) { init(ch_rng(0, 0)); }
      text(mem_pool& m, const ch_rng& r)  : mp(m) { init(r); }
      text(const text& t)  : mp(t.mp) { init(ch_rng(t)); }
      ~text() { mp.deallocate(buf, resd); }

      text& operator=(const text& t) { return *this=ch_rng(t); }
      text& operator=(const ch_rng& r);

      text& append(const ch_rng& r);
      text& append(char byte);

      int size() const { return used; }
      int capacity() const { return resd; }

      void reserve(int sz);
      void clear() { used=0; }
      void uninitialized_resize(int sz);

      char* begin() { return buf; }
      const char* begin() const { return buf; }
      char* end() { return buf+used; }
      const char* end() const { return buf+used; }

      const char* c_str() const;

      char& front() { return *begin(); }
      char front() const { return *begin(); }
      char& back() { return *(end()-1); }
      char back() const { return *(end()-1); }

      char* find(char ch);
      const char* find(char ch) const;
      char* rfind(char ch);
      const char* rfind(char ch) const;

      char* find(const ch_rng& r);
      const char* find(const ch_rng& r) const;
      char* rfind(const ch_rng& r);
      const char* rfind(const ch_rng& r) const;

      bool starts(const ch_rng& r) const;
      bool ends(const ch_rng& r) const;
};

namespace detail {  // ::ders::detail

int compare(const ch_rng& r1, const ch_rng& r2);

}  // namespace ::ders::detail

inline sh_text nt(mem_pool& mp)
{ return nt(mp, ch_rng(0, 0)); }

inline sh_text nt(mem_pool& mp, const char* str)
{ return nt(mp, ch_rng(str)); }

inline sh_text nt(mem_pool& mp, const char* b, const char* e)
{ return nt(mp, ch_rng(b, e)); }

inline sh_text nt(mem_pool& mp, const text& t)
{ return nt(mp, ch_rng(t)); }

inline sh_text nt(mem_pool& mp, const sh_text& sht)
{ return nt(mp, ch_rng(sht)); }

inline bool operator!=(const ch_rng& r1, const ch_rng& r2)
{ return !(r1==r2); }

inline bool operator<(const ch_rng& r1, const ch_rng& r2)
{ return detail::compare(r1, r2)<0; }

inline bool operator<=(const ch_rng& r1, const ch_rng& r2)
{ return detail::compare(r1, r2)<=0; }

inline bool operator>(const ch_rng& r1, const ch_rng& r2)
{ return detail::compare(r1, r2)>0; }

inline bool operator>=(const ch_rng& r1, const ch_rng& r2)
{ return detail::compare(r1, r2)>=0; }

inline ch_rng::ch_rng(const char* b, const char* e) : beg(b), end(e)
{ assert(beg!=0 && beg<=end || beg==0 && end==0); }

inline ch_rng::ch_rng(const text& t) : beg(t.begin()), end(t.end())
{ assert(beg!=0 && beg<=end); }

inline ch_rng::ch_rng(const sh_text& sht) : beg(sht->begin()), end(sht->end())
{ assert(beg!=0 && beg<=end); }

inline const char* text::find(char ch) const
{ return const_cast<text*>(this)->find(ch); }

inline const char* text::rfind(char ch) const
{ return const_cast<text*>(this)->rfind(ch); }

inline const char* text::find(const ch_rng& r) const
{ return const_cast<text*>(this)->find(r); }

inline const char* text::rfind(const ch_rng& r) const
{ return const_cast<text*>(this)->rfind(r); }

}  // namespace ::ders

#endif  // __DERS__TEXT_HPP__

