/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Defines interfaces @link ders::thread thread @endlink, @link ders::mutex
 * mutex @endlink, @link ders::cond_var cond_var @endlink, @link
 * ders::barrier_listener barrier_listener @endlink and
 * classes @link ders::auto_mutex auto_mutex @endlink, @link ders::barrier
 * barrier @endlink and
 * functions new_thread(), new_mutex(), new_long_mutex(), new_cond_var(),
 * new_barrier().
 */

#ifndef __DERS__THREAD_HPP__
#define __DERS__THREAD_HPP__

#include <ders/config.hpp>
#include <ders/sh_ptr.hpp>

namespace ders {  // ::ders

/**
 * Interface that defines implementation independent %thread.
 */
class thread : virtual public destroyable {
 public:
      virtual bool is_running()=0;
      virtual void join()=0;
};

typedef sh_ptr<thread> sh_thread;

/**
 * Creates new thread.
 * @param mp mem_pool to use.
 * @param start thread start function.
 * @param arg thread start function argument.
 * @return created thread object.
 */
sh_thread new_thread(mem_pool& mp, void (*start)(void*), void* arg);

/**
 * Interface that defines implementation independent %mutex.
 */
class mutex : virtual public destroyable {
 public:
      virtual void lock()=0;
      virtual void unlock()=0;
};

typedef sh_ptr<mutex> sh_mutex;

/**
 * Creates new mutex object for usual short term locking.
 * @param mp mem_pool to use.
 * @return created object.
 */
sh_mutex new_mutex(mem_pool& mp);

/**
 * Creates new mutex object for long term locking.
 * @param mp mem_pool to use.
 * @return created object.
 */
sh_mutex new_long_mutex(mem_pool& mp);

/**
 * Helper class for exception-safe %mutex locking.
 */
class auto_mutex {
      mutex& mtx;

      auto_mutex(const auto_mutex&);
      auto_mutex& operator=(const auto_mutex&);

 public:
      auto_mutex(mutex& m) : mtx(m) { mtx.lock(); }
      ~auto_mutex() { mtx.unlock(); }
};

/**
 * Interface that defines implementation independent condition variable.
 */
class cond_var : virtual public destroyable {
 public:
      static const long infinite=-1;

      void wait() { wait(infinite); }
      virtual void wait(long timeout)=0;
      virtual void notify()=0;
      virtual void notify_all()=0;
};

typedef sh_ptr<cond_var> sh_cond_var;

/**
 * Creates new cond_var object.
 * @param mp mem_pool to use.
 * @param mtx asociated mutex.
 * @return created object.
 */
sh_cond_var new_cond_var(mem_pool& mp, mutex& mtx);

class barrier;

/**
 * Interface for classes that listen for %on_release() barrier events.
 */
class barrier_listener : virtual public destroyable {
 public:
      /** action to take */
      enum action { release_one_false, release_all_false, release_all_true };

      virtual action on_release(barrier& bar)=0;
};

/**
 * Represents implementation independent %thread %barrier.
 */
class barrier {
      sh_mutex mtx;
      sh_cond_var cv1, cv2;
      int maxCnt, wtCnt, relCnt;
      bool ret;
      barrier_listener* blnr;

 public:
      barrier(mem_pool& mp, int cnt);
      ~barrier();

      int inc_count();
      int dec_count();
      int get_count();

      bool wait();
      void release_one();
      void release_all();

      barrier_listener* set_listener(barrier_listener* bl);
};

typedef sh_ptr<barrier> sh_barrier;

/**
 * Creates new barrier object.
 * @param mp mem_pool to use.
 * @param cnt initial threads count.
 * @return created object.
 */
sh_barrier new_barrier(mem_pool& mp, int cnt);

namespace detail {  // ::ders::detail

struct thr_core {
       void (*strt)(void*);
       void* arg;

       mem_pool ownmp;
       sh_mutex mtx;
       sh_cond_var cv;
       int refs;
       bool rng;

       thr_core(void (*st)(void*), void* ar);

       void inc_refs();
       void dec_refs(bool parent);
       void clear_rng();
       void start();

       bool is_running();
       void join();
};

struct tc_ptr {
       bool parent;
       thr_core* ptr;

       tc_ptr(bool pr, thr_core* pt) : parent(pr), ptr(pt) { ptr->inc_refs(); }
       ~tc_ptr() { ptr->dec_refs(parent); }

       thr_core* operator->() { return ptr; }
};

}  // namespace ::ders::detail

}  // namespace ::ders

#endif  // __DERS__THREAD_HPP__

