/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Implements classes @link ders::buf_reader buf_reader @endlink, @link
 * ders::buf_writer buf_writer @endlink.
 */

#include <ders/buf_rdwr.hpp>
#include <string.h>

namespace ders {  // ::ders

int buf_reader::read_impl()
{
 assert(it==lim);

 int rc=rd.read(buf, bsz);
 switch (rc) {
        case -1:
        case  0: return -1;
 }

 it=1;
 lim=rc;

 return buf[0];
}

buf_reader::buf_reader(mem_pool& m, readable& rdbl, int bufsz) :
  mp(m), rd(rdbl), bsz(bufsz)
{
 assert(bsz>0);

 buf=static_cast<unsigned char*>(mp.allocate(bsz));
 reset();
}

int buf_reader::read(void* b, int n)
{
 assert(n>=0);

 if (n==0) return 0;

 if (it==lim) {
    int rc=rd.read(buf, bsz);
    switch (rc) {
           case -1: return -1;
           case  0: return 0;
    }

    it=0;
    lim=rc;
 }

 if (n<=lim-it) {
    memcpy(b, buf, n);
    it+=n;

    return n;
 }

 // n>lim-it
 int rdn=lim-it;
 memcpy(b, buf, rdn);
 reset();

 b=static_cast<unsigned char*>(b)+rdn;
 n-=rdn;

 int rc=rd.read(b, n);
 return (rc!=-1) ? rdn+rc : rdn;
}

int buf_reader::read(sh_text& buf, int n)
{
 assert(n>=0);

 buf->clear();
 buf->reserve(n);

 int rc=read(buf->begin(), n);
 assert(rc<=n);
 if (rc!=-1) buf->uninitialized_resize(rc);

 return rc;
}

bool buf_reader::read_line(sh_text& line, bool noendl)
{
 line->clear();

 for (int ch; ; ) {
     if ((ch=read())==-1) return line->size()!=0;

     if (ch!='\n') line->append(ch);
     else {
          if (!noendl) line->append(ch);
          else {
               if (line->size()>0 && line->back()=='\r')
                  line->uninitialized_resize(line->size()-1);
          }

          break;
     }
 }

 return true;
}

int buf_writer::write_impl(char ch)
{
 assert(it==bsz);

 int rc=wr.write(buf, it);
 switch (rc) {
        case -1:
        case  0: return -1;
 }

 if (it==rc) {
    it=1;
    return buf[0]=ch;
 }

 it-=rc;
 memmove(buf, buf+rc, it);
 return buf[it++]=ch;
}

buf_writer::buf_writer(mem_pool& m, writable& wrbl, int bufsz) :
  mp(m), wr(wrbl), bsz(bufsz)
{
 assert(bsz>0);

 buf=static_cast<unsigned char*>(mp.allocate(bsz));
 reset();
}

buf_writer::~buf_writer()
{
 flush();
 mp.deallocate(buf, bsz);
}

int buf_writer::write(const void* b, int n)
{
 assert(n>=0);

 if (n==0) return 0;

 int wrn=bsz-it;
 if (n<=wrn) {
    memcpy(buf+it, b, n);
    it+=n;

    return n;
 }

 // n>wrn
 memcpy(buf+it, b, wrn);
 it=bsz;

 if (!flush()) return wrn;

 b=static_cast<const char*>(b)+wrn;
 n-=wrn;

 int rc=wr.write(b, n);
 return (rc!=-1) ? wrn+rc : wrn;
}

bool buf_writer::flush()
{
 if (it==0) return true;

 int rc=wr.write(buf, it);
 if (rc==-1) return false;

 memmove(buf, buf+rc, it-rc);
 it-=rc;

 return it==0;
}

}  // namespace ::ders

