/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Implements functions error_message(), convert_errno() and
 * exception class @link ders::ErrorException ErrorException @endlink.
 */

#include <ders/error.hpp>
#include <errno.h>
#include <ders/text_buf.hpp>

namespace {

const char* error_strs[]={
  "No error",
  "Argument list too long",
  "Permission denied",
  "Address in use",
  "Address not available",
  "Address family not supported",
  "Resource unavailable, try again",
  "Connection already in progress",
  "Bad file descriptor",
  "Bad message",
  "Device or resource busy",
  "Operation canceled",
  "No child processes",
  "Connection aborted",
  "Connection refused",
  "Connection reset",
  "Resource deadlock would occur",
  "Destination address required",
  "Mathematics argument out of domain of function",
  "Reserved (EDQUOT)",
  "File exists",
  "Bad address",
  "File too large",
  "Host is unreachable",
  "Identifier removed",
  "Illegal byte sequence",
  "Operation in progress",
  "Interrupted function",
  "Invalid argument",
  "I/O error",
  "Socket is connected",
  "Is a directory",
  "Too many levels of symbolic links",
  "Too many open files",
  "Too many links",
  "Message too large",
  "Reserved (EMULTIHOP)",
  "Filename too long",
  "Network is down ",
  "Connection aborted by network",
  "Network unreachable",
  "Too many files open in system",
  "No buffer space available",
  "No message is available on the STREAM head read queue",
  "No such device",
  "No such file or directory",
  "Executable file format error",
  "No locks available",
  "Reserved (ENOLINK)",
  "Not enough space",
  "No message of the desired type",
  "Protocol not available",
  "No space left on device",
  "No STREAM resources",
  "Not a STREAM",
  "Function not supported",
  "The socket is not connected",
  "Not a directory",
  "Directory not empty",
  "Not a socket",
  "Not supported",
  "Inappropriate I/O control operation",
  "No such device or address",
  "Operation not supported on socket",
  "Value too large to be stored in data type",
  "Operation not permitted",
  "Broken pipe",
  "Protocol error",
  "Protocol not supported",
  "Protocol wrong type for socket",
  "Result too large",
  "Read-only file system",
  "Invalid seek",
  "No such process",
  "Reserved (ESTALE)",
  "Stream ioctl() timeout",
  "Connection timed out",
  "Text file busy",
  "Unknown error",
  "Operation would block",
  "Cross-device link"
};

}  // unnamed

namespace ders {  // ::ders

sh_text error_message(mem_pool& mp, error err)
{
 assert(unsigned(err)<sizeof(error_strs)/sizeof(*error_strs));
 return nt(mp, error_strs[err]);
}

error convert_errno(int val)
{
 switch (val) {
        case 0: return ezero;
#ifdef E2BIG
        case E2BIG: return e2big;
#endif
#ifdef EACCES
        case EACCES: return eacces;
#endif
#ifdef EADDRINUSE
        case EADDRINUSE: return eaddrinuse;
#endif
#ifdef EADDRNOTAVAIL
        case EADDRNOTAVAIL: return eaddrnotavail;
#endif
#ifdef EAFNOSUPPORT
        case EAFNOSUPPORT: return eafnosupport;
#endif
#ifdef EAGAIN
        case EAGAIN: return eagain;
#endif
#ifdef EALREADY
        case EALREADY: return ealready;
#endif
#ifdef EBADF
        case EBADF: return ebadf;
#endif
#ifdef EBADMSG
        case EBADMSG: return ebadmsg;
#endif
#ifdef EBUSY
        case EBUSY: return ebusy;
#endif
#ifdef ECANCELED
        case ECANCELED: return ecanceled;
#endif
#ifdef ECHILD
        case ECHILD: return echild;
#endif
#ifdef ECONNABORTED
        case ECONNABORTED: return econnaborted;
#endif
#ifdef ECONNREFUSED
        case ECONNREFUSED: return econnrefused;
#endif
#ifdef ECONNRESET
        case ECONNRESET: return econnreset;
#endif
#ifdef EDEADLK
        case EDEADLK: return edeadlk;
#endif
#ifdef EDESTADDRREQ
        case EDESTADDRREQ: return edestaddrreq;
#endif
#ifdef EDOM
        case EDOM: return edom;
#endif
#ifdef EDQUOT
        case EDQUOT: return edquot;
#endif
#ifdef EEXIST
        case EEXIST: return eexist;
#endif
#ifdef EFAULT
        case EFAULT: return efault;
#endif
#ifdef EFBIG
        case EFBIG: return efbig;
#endif
#ifdef EHOSTUNREACH
        case EHOSTUNREACH: return ehostunreach;
#endif
#ifdef EIDRM
        case EIDRM: return eidrm;
#endif
#ifdef EILSEQ
        case EILSEQ: return eilseq;
#endif
#ifdef EINPROGRESS
        case EINPROGRESS: return einprogress;
#endif
#ifdef EINTR
        case EINTR: return eintr;
#endif
#ifdef EINVAL
        case EINVAL: return einval;
#endif
#ifdef EIO
        case EIO: return eio;
#endif
#ifdef EISCONN
        case EISCONN: return eisconn;
#endif
#ifdef EISDIR
        case EISDIR: return eisdir;
#endif
#ifdef ELOOP
        case ELOOP: return eloop;
#endif
#ifdef EMFILE
        case EMFILE: return emfile;
#endif
#ifdef EMLINK
        case EMLINK: return emlink;
#endif
#ifdef EMSGSIZE
        case EMSGSIZE: return emsgsize;
#endif
#ifdef EMULTIHOP
        case EMULTIHOP: return emultihop;
#endif
#ifdef ENAMETOOLONG
        case ENAMETOOLONG: return enametoolong;
#endif
#ifdef ENETDOWN
        case ENETDOWN: return enetdown;
#endif
#ifdef ENETRESET
        case ENETRESET: return enetreset;
#endif
#ifdef ENETUNREACH
        case ENETUNREACH: return enetunreach;
#endif
#ifdef ENFILE
        case ENFILE: return enfile;
#endif
#ifdef ENOBUFS
        case ENOBUFS: return enobufs;
#endif
#ifdef ENODATA
        case ENODATA: return enodata;
#endif
#ifdef ENODEV
        case ENODEV: return enodev;
#endif
#ifdef ENOENT
        case ENOENT: return enoent;
#endif
#ifdef ENOEXEC
        case ENOEXEC: return enoexec;
#endif
#ifdef ENOLCK
        case ENOLCK: return enolck;
#endif
#ifdef ENOLINK
        case ENOLINK: return enolink;
#endif
#ifdef ENOMEM
        case ENOMEM: return enomem;
#endif
#ifdef ENOMSG
        case ENOMSG: return enomsg;
#endif
#ifdef ENOPROTOOPT
        case ENOPROTOOPT: return enoprotoopt;
#endif
#ifdef ENOSPC
        case ENOSPC: return enospc;
#endif
#ifdef ENOSR
        case ENOSR: return enosr;
#endif
#ifdef ENOSTR
        case ENOSTR: return enostr;
#endif
#ifdef ENOSYS
        case ENOSYS: return enosys;
#endif
#ifdef ENOTCONN
        case ENOTCONN: return enotconn;
#endif
#ifdef ENOTDIR
        case ENOTDIR: return enotdir;
#endif
#ifdef ENOTEMPTY
        case ENOTEMPTY: return enotempty;
#endif
#ifdef ENOTSOCK
        case ENOTSOCK: return enotsock;
#endif
#if defined(ENOTSUP) && ENOTSUP!=EOPNOTSUPP
        case ENOTSUP: return enotsup;
#endif
#ifdef ENOTTY
        case ENOTTY: return enotty;
#endif
#ifdef ENXIO
        case ENXIO: return enxio;
#endif
#ifdef EOPNOTSUPP
        case EOPNOTSUPP: return eopnotsupp;
#endif
#ifdef EOVERFLOW
        case EOVERFLOW: return eoverflow;
#endif
#ifdef EPERM
        case EPERM: return eperm;
#endif
#ifdef EPIPE
        case EPIPE: return epipe;
#endif
#ifdef EPROTO
        case EPROTO: return eproto;
#endif
#ifdef EPROTONOSUPPORT
        case EPROTONOSUPPORT: return eprotonosupport;
#endif
#ifdef EPROTOTYPE
        case EPROTOTYPE: return eprototype;
#endif
#ifdef ERANGE
        case ERANGE: return erange;
#endif
#ifdef EROFS
        case EROFS: return erofs;
#endif
#ifdef ESPIPE
        case ESPIPE: return espipe;
#endif
#ifdef ESRCH
        case ESRCH: return esrch;
#endif
#ifdef ESTALE
        case ESTALE: return estale;
#endif
#ifdef ETIME
        case ETIME: return etime;
#endif
#ifdef ETIMEDOUT
        case ETIMEDOUT: return etimedout;
#endif
#ifdef ETXTBSY
        case ETXTBSY: return etxtbsy;
#endif
#if defined(EWOULDBLOCK) && EWOULDBLOCK!=EAGAIN
        case EWOULDBLOCK: return ewouldblock;
#endif
#ifdef EXDEV
        case EXDEV: return exdev;
#endif
        default: return eunkn;
 }
}

shException newErrorException(mem_pool& mp, const FileLine& loc, const ch_rng&
  msg, error err)
{
 return newErrorException(loc, msg, err, shException(mp, 0));
}

shException newErrorException(const FileLine& loc, const ch_rng& msg, error err,
  shException nest)
{
 mp_newbuf<ErrorException> buf(nest.pool());
 return shException(buf.pool(), buf.rls(::new(buf.get()) ErrorException(loc,
   msg, err, nest)));
}

ErrorException::ErrorException(const FileLine& loc, const ch_rng& msg, error
  err, shException nest) : Exception(loc, msg, nest), errorCode(err)
{
}

sh_text ErrorException::getClassName() const
{
 return nt(message.pool(), "ders::ErrorException");
}

sh_text ErrorException::toText() const
{
 return text_buf(getClassName())+" ["+location.file+":"+location.line+
   "], errorCode="+errorCode+":\""+error_message(message.pool(), errorCode)+
   "\", message: "+message;
}

}  // namespace ::ders

