/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * POSIX implementation of functions new_dir(), make_dir().
 */

#include <ders/dir.hpp>
#include <dirent.h>
#include <errno.h>
#include <sys/stat.h>
#include <ders/file.hpp>
#include <ders/text_buf.hpp>

namespace {

using namespace ders;

struct pos_dir : public dir {
       sh_text dn;
       DIR* dr;

       pos_dir(sh_text dname) : dn(dname), dr(0) {}
       ~pos_dir() { if (dr) closedir(dr); }

       virtual bool find_next(entry&);
       virtual sh_text full_name(const entry& dent);

       virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
};

bool pos_dir::find_next(entry& dent)
{
 mem_pool& mp=dn.pool();

 if (!dr) {
    if ( !(dr=opendir(dn->size() ? dn->c_str() : ".")) ) {
       throw newFileErrorException(mp, _FLINE_, "Can't find first file",
         convert_errno(errno), dn);
    }
 }

 dirent *ent=readdir(dr);
 if (!ent) return false;

 *dent.name=ent->d_name;
 dent.isdir=false;
 dent.size=0;

 struct stat statbuf;
 if (stat(pos_dir::full_name(dent)->c_str(), &statbuf)==0) {
    dent.isdir=statbuf.st_mode&S_IFDIR;
    dent.size=statbuf.st_size;
 }

 return true;
}

sh_text pos_dir::full_name(const entry& dent)
{
 text_buf buf(dn);
 if (dn->size() && dn->back()!='/') buf+'/';

 return buf+dent.name;
}

}  // unnamed

namespace ders {  // ::ders

sh_dir new_dir(mem_pool& mp, const ch_rng& dname)
{
 mp_newbuf<pos_dir> buf(mp);
 return sh_dir(buf.pool(), buf.rls(::new(buf.get()) pos_dir(nt(mp, dname))));
}

error make_dir(mem_pool& mp, const ch_rng& dname)
{
 return (mkdir(text(mp, dname).c_str(), S_IRWXU|S_IRWXG|S_IRWXO)==-1) ?
   convert_errno(errno) : ezero;
}

}  // namespace ::ders

