/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Implements struct @link ders::ch_rng ch_rng @endlink and
 * class @link ders::text text @endlink and
 * set of functions for ch_rng comparison.
 */

#include <ders/text.hpp>
#include <string.h>
#include <ders/mem.hpp>

namespace ders {  // ::ders

namespace detail {  // ::ders::detail

int compare(const ch_rng& r1, const ch_rng& r2)
{
 int sz1=r1.size(), sz2=r2.size();

 int diff=memcmp(r1.beg, r2.beg, (sz1<sz2) ? sz1 : sz2);
 return (diff) ? diff : sz1-sz2;
}

}  // namespace ::ders::detail

sh_text nt(mem_pool& mp, const ch_rng& r)
{
 mp_newbuf<text> buf(mp);
 return sh_text(buf.pool(), buf.rls(::new(buf.get()) text(buf.pool(), r)));
}

bool operator==(const ch_rng& r1, const ch_rng& r2)
{
 int sz1=r1.size(), sz2=r2.size();
 return sz1==sz2 && memcmp(r1.beg, r2.beg, sz1)==0;
}

ch_rng::ch_rng(const char* str)
{
 assert(str!=0);
 beg=str;
 end=str+strlen(str);
}

void text::init(const ch_rng& r)
{
 used=r.size();
 resd=used+1;

 buf=static_cast<char*>(mp.allocate(resd));
 memcpy(buf, r.beg, used);
}

text& text::operator=(const ch_rng& r)
{
 int sz=r.size();
 reserve(sz);

 memcpy(buf, r.beg, sz);
 used=sz;

 return *this;
}

text& text::append(const ch_rng& r)
{
 int sz=r.size(), newsz=used+sz;
 reserve(newsz);

 memcpy(buf+used, r.beg, sz);
 used=newsz;

 return *this;
}

text& text::append(char byte)
{
 reserve(used+1);
 buf[used++]=byte;

 return *this;
}

void text::reserve(int sz)
{
 assert(sz>=0);

 sz++;
 if (sz<=resd) return;
 if (resd*2>sz) sz=resd*2;

 char* newbuf=static_cast<char*>(mp.allocate(sz));
 memcpy(newbuf, buf, used);

 mp.deallocate(buf, resd);
 buf=newbuf;
 resd=sz;
}

void text::uninitialized_resize(int sz)
{
 assert(sz>=0);

 reserve(sz);
 used=sz;
}

const char* text::c_str() const
{
 assert(resd>used);
 buf[used]=0;

 return buf;
}

char* text::find(char ch)
{
 char* ptr=(char*)mem_find(begin(), size(), ch);
 return ptr ? ptr : end();
}

char* text::rfind(char ch)
{
 char* ptr=(char*)mem_rev_find(begin(), size(), ch);
 return ptr ? ptr : end();
}

char* text::find(const ch_rng& r)
{
 char* ptr=(char*)mem_find(begin(), size(), r.beg, r.size());
 return ptr ? ptr : end();
}

char* text::rfind(const ch_rng& r)
{
 char* ptr=(char*)mem_rev_find(begin(), size(), r.beg, r.size());
 return ptr ? ptr : end();
}

bool text::starts(const ch_rng& r) const
{
 int sz=r.size();
 return size()>=sz && memcmp(begin(), r.beg, sz)==0;
}

bool text::ends(const ch_rng& r) const
{
 int sz=r.size();
 return size()>=sz && memcmp(end()-sz, r.beg, sz)==0;
}

}  // namespace ::ders

