/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Internal Exception test.
 */

#include <vector>
#include <ders/file.hpp>
#include <ders/text_buf.hpp>

using namespace std;
using namespace ders;

shException newExampleException(mem_pool& mp, const FileLine& loc, const ch_rng&
  msg);
shException newExampleException(const FileLine& loc, const ch_rng& msg,
  shException nest);

class ExampleException : public Exception {
      friend shException newExampleException(const FileLine& loc, const ch_rng&
        msg, shException nest);

      ExampleException(const ExampleException&);
      ExampleException& operator=(const ExampleException&);

 protected:
      ExampleException(const FileLine& loc, const ch_rng& msg, shException
        nest);

 public:
      virtual sh_text getClassName() const;

      virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
};

shException newExampleException(mem_pool& mp, const FileLine& loc, const ch_rng&
  msg)
{
 return newExampleException(loc, msg, shException(mp, 0));
}

shException newExampleException(const FileLine& loc, const ch_rng& msg,
  shException nest)
{
 mp_newbuf<ExampleException> buf(nest.pool());
 return shException(buf.pool(), buf.rls(::new(buf.get()) ExampleException(loc,
   msg, nest)));
}

ExampleException::ExampleException(const FileLine& loc, const ch_rng& msg,
  shException nest) : Exception(loc, msg, nest)
{
}

sh_text ExampleException::getClassName() const
{
 return nt(message.pool(), "ExampleException");
}

void work(mem_pool& mp);
void f(mem_pool& mp, int i);
void g(mem_pool& mp, int i);

int main()
{
 mem_pool mp;
 file err(mp, fd::err);
 file out(mp, fd::out);

 shException exc(mp, 0);
 try {
     work(mp);
     return 0;
 }
 catch (shException she) { exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 ExitMsgException* em=exc->is<ExitMsgException>();
 if (em) {
    if (em->message->size())
       (em->exitCode ? err : out).write(text_buf(em->message)+'\n');
    return em->exitCode;
 }

 err.write(text_buf(mp)+"Uncaught exception:\n"+toTextAll(exc)+'\n');
 return 1;
}

void work(mem_pool& mp)
{
 file out(mp, fd::out);

 for (int i=1; i<=5; i++) {
     shException exc(mp, 0);
     try {
         if (i!=4) f(mp, i);
         else throw 4;
     }
     catch (shException she) {
           if (she->is<MsgException>()) throw; else exc=she;
     }
     catch (...) { exc=recatchException(mp, _FLINE_); }

     if (exc.get())
        out.write(text_buf(mp)+"\tException #"+i+":\n"+toTextAll(exc)+"\n\n");
 }
}

void f(mem_pool& mp, int i)
{
 shException exc(mp, 0);
 try {
     g(mp, i);
     return;
 }
 catch (shException she) { if (she->is<MsgException>()) throw; else exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 throw newException(_FLINE_, "Problems in f()", exc);
}

void g(mem_pool& mp, int i)
{
 shException exc(mp, 0);
 try {
     switch (i) {
            case 1: {
                 throw newExampleException(mp, _FLINE_, "Hello from g()");
            }
            case 2: {
                 vector<int> v;
                 v.at(0);  // std::out_of_range
            }
            case 3: {
                 throw 3;
            }
            case 5: {
                 throw newExitMsgException(mp, _FLINE_,
                   "Non-empty exit message", 0);
            }
     }

     return;
 }
 catch (shException she) { if (she->is<MsgException>()) throw; else exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 throw newException(_FLINE_, "Problems in g()", exc);
}

