/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Internal thread_pool test.
 */

#include <algorithm>
#include <stdlib.h>
#include <ders/file.hpp>
#include <ders/text_buf.hpp>
#include <ders/thread_pool.hpp>

using namespace ders;
using namespace std;

const int N=1000000;

struct test {
       mem_pool& mp;
       vector<int> in;
       int inc;
       vector<int> out;

       test(mem_pool& m);

       bool get(int& val);
       void put(int val);
       void check(file& err);
};

struct test_io : object_io {
       int val;

       virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
       virtual bool read_from(data_in& in) { return in.read_int(val); }
       virtual void write_to(data_out& out) const { out.write_int(val); }
};

struct test_task : task {
       struct inv_data {
              test& tst;
              int val;
              bool rc;

              inv_data(test& t) : tst(t) {}
       };

       int num;
       test& tst;

       test_task(test& t) : num(0), tst(t) {}

       static void get(void* arg);
       static void put(void* arg);

       virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
       virtual void proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers&
        to);

       void next() { num++; }
       bool read(data_queue* dq, int& val, bool wait=true);
       void write(data_queue* dq, int val);
};

test::test(mem_pool& m) : mp(m), inc(0)
{
 in.reserve(N);
 for (int i=0; i<N; i++) in.push_back(i);
}

bool test::get(int& val)
{
 if (inc>=int(in.size())) return false;

 val=in[inc++];
 return true;
}

void test::put(int val)
{
 hard_assert(out.size()<in.size());
 out.push_back(val);
}

void test::check(file& err)
{
 hard_assert(inc==int(in.size()));
 hard_assert(out.size()==in.size());

 sort(out.begin(), out.end());

 for (int i=0, end=out.size(); i<end; i++) {
     if (out[i]!=in[i]) {
        err.write(text_buf(mp)+"["+i+"]: out="+out[i]+" in="+in[i]+"\n");
        hard_assert(false);
     }
 }
}

void test_task::get(void* arg)
{
 inv_data& id=*static_cast<inv_data*>(arg);
 id.rc=id.tst.get(id.val);
}

void test_task::put(void* arg)
{
 inv_data& id=*static_cast<inv_data*>(arg);
 id.tst.put(id.val);
}

void test_task::proc(mem_pool&, const dq_vec& dqv, void*, task_opers& to)
{
 switch (num) {
        case 0: {
             inv_data id(tst);
             for (;;) {
                 to.invoke(get, &id);
                 if (!id.rc) break;

                 write(dqv[0], id.val);
             }

             break;
        }
        case 1: {
             for (int i=0; ; i++) {
                 int val;
                 if (!read(dqv[0], val)) break;

                 write(dqv[(i&1) ? 1 : 2], val);
             }

             break;
        }
        case 2: {
             inv_data id(tst);
             for (int i=0; ; i++) {
                 data_queue *dq1, *dq2;
                 if (i&1) { dq1=dqv[1], dq2=dqv[2]; }
                 else  { dq1=dqv[2], dq2=dqv[1]; }

                 if (!read(dq1, id.val, false) && !read(dq2, id.val)) break;

                 to.invoke(put, &id);
             }

             break;
        }
        case 3: {
             int sz=dqv.size();

             vector<int> is;
             for (int i=0; i<sz; i++) is.push_back(i);

             test_io io;
             do {
                for (int i=0; i<sz; i++) {
                    data_queue::result rc=dqv[is[i]]->read_obj(io);
                    hard_assert(rc==data_queue::empt);
                }
             } while (next_permutation(is.begin(), is.end()));

             break;
        }
        case 4: {
             test_io io;
             for (int i=0, end=dqv.size(); i<end; i++) {
                 dqv[i]->set_intr(true);
                 hard_assert(dqv[i]->is_intr());

                 data_queue::result rc=dqv[i]->read_obj(io);
                 hard_assert(rc==data_queue::intr);

                 rc=dqv[i]->write_obj(io);
                 hard_assert(rc==data_queue::intr);
             }

             break;
        }
        default: hard_assert(false);
 }
}

bool test_task::read(data_queue* dq, int& val, bool wait)
{
 test_io io;

 data_queue::result rc=dq->read_obj(io, wait);
 switch (rc) {
        case data_queue::ok: {
             val=io.val;
             return true;
        }
        case data_queue::empt: {
             return false;
        }
        default: hard_assert(false);
 }

 // never get here
 return false;
}

void test_task::write(data_queue* dq, int val)
{
 test_io io;
 io.val=val;

 dq->write_obj(io);
}

int main(int argc, char** argv)
{
 mem_pool mp;
 file err(mp, fd::err);

 if (argc!=2) {
    err.write(text_buf(mp)+argv[0]+" num_threads");
    return 1;
 }

 int numThr=atoi(argv[1]);
 if ( !(numThr>=1 && numThr<=100) ) {
    err.write("num_threads must be in [1, 100]");
    return 1;
 }

 sh_data_queue dq1=new_data_queue(mp), dq2=new_data_queue(mp), dq3=
   new_data_queue(mp);

 dq_vec dqv;
 dqv.push_back(dq1.get());
 dqv.push_back(dq2.get());
 dqv.push_back(dq3.get());

 test tst(mp);
 test_task tsk(tst);
 sh_thread_pool tp=(numThr>1) ? new_thread_pool(mp, numThr-1) : new_thread_pool(
   mp);

 for (int i=0; i<5; i++) {
     if (i) tsk.next();
     tp->exec(tsk, dqv);
 }

 tst.check(err);

 return 0;
}

