/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Example of file performance.
 */

#include <memory>
#include <vector>
#include <stdio.h>
#include <time.h>
#include <ders/file.hpp>
#include <ders/text_buf.hpp>
#include <ders/thread.hpp>

using namespace std;
using namespace ders;

const int BUF_SIZE=64*1024;

struct MainData {
       const char* fname;
       MainData(const char* fn) : fname(fn) {}
};

struct ThreadData {
       MainData* md;
       int n;

       ThreadData(MainData* md_, int n_) : md(md_), n(n_) {}
};

void start_std(void* arg)
{
 ThreadData* td=(ThreadData*)arg;
 auto_ptr<ThreadData> guard(td);

 mem_pool mp;
 file err(mp, fd::err);

 FILE* fin=fopen(td->md->fname, "rb");
 if (!fin) {
    err.write(text_buf(mp)+"Can't open "+td->md->fname+'\n');
    return;
 }

 sh_text oname=text_buf(mp)+td->md->fname+'.'+td->n;
 FILE* fout=fopen(oname->c_str(), "wb");
 if (!fout) {
    err.write(text_buf(mp)+"Can't create "+oname+'\n');
    fclose(fin);
    return;
 }

 setvbuf(fin, 0, _IOFBF, BUF_SIZE);
 setvbuf(fout, 0, _IOFBF, BUF_SIZE);

 for (int ch; (ch=fgetc(fin))!=EOF; )
     fputc(ch, fout);

 fclose(fout);
 fclose(fin);
}

void start_ders(void* arg)
{
 ThreadData* td=(ThreadData*)arg;
 auto_ptr<ThreadData> guard(td);

 mem_pool mp;
 file err(mp, fd::err);

 file fin(mp);
 if (!fin.open(td->md->fname, file::rdo, 0)) {
    err.write(text_buf(mp)+"Can't open "+td->md->fname+'\n');
    return;
 }

 sh_text oname=text_buf(mp)+td->md->fname+'.'+td->n;
 file fout(mp);
 if (!fout.open(oname, file::wro, file::crt|file::trnc)) {
    err.write(text_buf(mp)+"Can't create "+oname+'\n');
    return;
 }

 buf_reader br(mp, fin, BUF_SIZE);
 buf_writer bw(mp, fout, BUF_SIZE);

 for (int ch; (ch=br.read())!=-1; )
     bw.write(ch);
}

int main(int argc, char** argv)
{
 mem_pool mp;
 file err(mp, fd::err);
 file out(mp, fd::out);

 if (argc!=4) {
    m1:
    err.write("main file num_threads std|ders");
    return 1;
 }

 int numThr=atoi(argv[2]);
 if ( !(numThr>=1 && numThr<=100) ) {
    err.write("num_threads must be in [1, 100]");
    return 1;
 }

 void (*start)(void*);
 if (strcmp(argv[3], "std")==0) start=start_std;
 else if (strcmp(argv[3], "ders")==0) start=start_ders;
 else goto m1;

 MainData md(argv[1]);

 clock_t c1=clock();

 vector<sh_thread> vthr;
 for (int i=0; i<numThr; i++)
     vthr.push_back(new_thread(mp, start, new ThreadData(&md, i)));
 for (int i=0; i<numThr; i++) vthr[i]->join();

 clock_t c2=clock();
 out.write(text_buf(mp)+numThr+'\t'+int(c2-c1)+'\t'+argv[3]+'\n');

 return 0;
}

