/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Main file of mtcksrc program.
 */

#include <stdlib.h>
#include <ders/dir.hpp>
#include <ders/file.hpp>
#include <ders/mem.hpp>
#include <ders/path.hpp>
#include <ders/text_buf.hpp>
#include <ders/thread_pool.hpp>
#include <ders/wldcrd_mtchr.hpp>
#include "msg.hpp"

namespace mtcksrc {  // ::mtcksrc

using namespace ders;
using namespace std;

struct CmdLineParser {
       int numThr;
       bool isDos;
       sh_text masks;

       CmdLineParser(mem_pool& mp) : numThr(2), masks(nt(mp)) {}

       void parse(int argc, char** argv);
};

struct MainTask : public task {
       bool dos;
       vector<sh_wldcrd_mtchr> vmr;
       vector<int> vns;

       MainTask(bool d, sh_text ms);

       virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
       virtual void proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers&
         to);

       void doFindFiles(mem_pool& mp, data_queue& dq, const FindFilesMsg& msg);
       void doCheckFile(mem_pool& mp, data_queue& dq, const CheckFileMsg& msg);
};

void CmdLineParser::parse(int, char** argv)
{
 const char* usage="mtcksrc [-nt2] d|u mask[,mask2...]";
 mem_pool& mp=masks.pool();

 char** it=argv;
 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);

 const ch_rng ntpar("-nt");
 if (text(mp, *it).starts(ntpar)) {
    int val=atoi(*it+ntpar.size());
    if ( !(val>=1 && val<=100) ) {
       throw newExitMsgException(mp, _FLINE_,
         "-ntN must be between -nt1 and -nt100", 1);
    }

    numThr=val;

    if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 }

 const ch_rng dupar(*it);
 if (dupar=="d") isDos=true;
 else if (dupar=="u") isDos=false;
 else throw newExitMsgException(mp, _FLINE_, usage, 1);

 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 *masks=*it;

 if (*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
}

MainTask::MainTask(bool d, sh_text ms) : dos(d)
{
 mem_pool& mp=ms.pool();

 for (char* beg=ms->begin(); beg<ms->end(); ) {
     char* end=static_cast<char*>(mem_find(beg, ms->end()-beg, ','));
     if (!end) end=ms->end();

     sh_text esc_msk(nt(mp));
     int numSep=0;

     for (const char* ptr=beg; ptr<end; ptr++) {
         if (*ptr==pathSepr) numSep++;

         if (*ptr=='\\') esc_msk->append(*ptr);
         esc_msk->append(*ptr);
     }

     vmr.push_back(new_wldcrd_mtchr(mp, esc_msk));
     vns.push_back(numSep);

     beg=end+1;
 }
}

void MainTask::proc(mem_pool& mp, const dq_vec& dqv, void*, task_opers&)
{
 data_queue& dq=*dqv[0];
 for (;;) {
     shException exc(mp, 0);
     try {
         for (MsgIO mio(mp, dq); ; ) {
             sh_ptr<Msg> msg=mio.read();
             if (!msg.get()) break;

             switch (msg->getType()) {
                    case Msg::FindFiles: {
                         doFindFiles(mp, dq, msg->to<FindFilesMsg>());
                         break;
                    }
                    case Msg::CheckFile: {
                         doCheckFile(mp, dq, msg->to<CheckFileMsg>());
                         break;
                    }
             }
         }

         return;
     }
     catch (shException she) { exc=she; }
     catch (...) { exc=recatchException(mp, _FLINE_); }

     file(mp, fd::err).write(text_buf(toTextAll(exc))+'\n');

     dq.set_intr(true);
     break;
 }
}

void MainTask::doFindFiles(mem_pool& mp, data_queue& dq, const FindFilesMsg&
  msg)
{
 MsgIO mio(mp, dq);
 sh_dir shd=new_dir(mp, msg.dirName);

 for (dir::entry dent(mp); shd->find_next(dent); ) {
     if (dent.name=="." || dent.name=="..")
        continue;

     sh_text fname=shd->full_name(dent);

     if (dent.isdir) {
        FindFilesMsg m(fname);
        mio.write(m);

        continue;
     }

     assert(vmr.size()==vns.size());
     for (int i=0, end=vmr.size(); i<end; i++) {
         char* ptr=fname->end()-1;
         for (int ns=0; ptr>=fname->begin(); ptr--)
             if (*ptr==pathSepr && ++ns>vns[i]) break;

         if (vmr[i]->match(ch_rng(ptr+1, fname->end()))) {
            CheckFileMsg m(fname);
            mio.write(m);

            break;
         }
     }
 }
}

void MainTask::doCheckFile(mem_pool& mp, data_queue&, const CheckFileMsg& msg)
{
 text_buf tout(mp);

 file fin(mp, msg.fileName, file::rdo, 0);
 buf_reader br(mp, fin, 64*1024);

 sh_text line(nt(mp));
 for (int num=1; br.read_line(line, false); num++) {
     assert(line->size()>0);
     if (line->back()=='\n') {
        if (line->ends("\r\n")) {
           if (!dos) tout+msg.fileName+':'+num+":dos\n";
           line->uninitialized_resize(line->size()-2);
        }
        else {
             if (dos) tout+msg.fileName+':'+num+":unix\n";
             line->uninitialized_resize(line->size()-1);
        }
     }

     if (line->find('\t')!=line->end()) tout+msg.fileName+':'+num+":tab\n";

     if (line->size()>80) tout+msg.fileName+':'+num+":long "+line->size()+'\n';

     if (line->size()>0 && line->back()==' ')
        tout+msg.fileName+':'+num+":ending space\n";
 }

 if (tout.size()) file(mp, fd::out).write(tout);
}

}  // namespace ::mtcksrc

int main(int argc, char** argv)
{
 using namespace ders;
 using namespace mtcksrc;

 mem_pool mp;
 file err(mp, fd::err);
 file out(mp, fd::out);

 shException exc(mp, 0);
 try {
     CmdLineParser clp(mp);
     clp.parse(argc, argv);

     MainTask mt(clp.isDos, clp.masks);

     sh_data_queue dq=new_data_queue(mp);

     MsgIO mio(mp, *dq);
     FindFilesMsg m(nt(mp, ""));
     mio.write(m);

     sh_thread_pool tp=(clp.numThr>1) ? new_thread_pool(mp, clp.numThr-1) :
       new_thread_pool(mp);
     tp->exec(mt, dq_vec(1, dq.get()));

     return (dq->is_intr()) ? 2 : 0;
 }
 catch (shException she) { exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 ExitMsgException* em=exc->is<ExitMsgException>();
 if (em) {
    if (em->message->size())
       (em->exitCode ? err : out).write(text_buf(em->message)+'\n');
    return em->exitCode;
 }

 err.write(text_buf(toTextAll(exc))+'\n');
 return 2;
}

