/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Main file of mtcnvsrc program.
 */

#include <stdlib.h>
#include <ders/buf_cache.hpp>
#include <ders/dir.hpp>
#include <ders/file.hpp>
#include <ders/mem.hpp>
#include <ders/path.hpp>
#include <ders/text_buf.hpp>
#include <ders/thread_pool.hpp>
#include <ders/wldcrd_mtchr.hpp>
#include "msg.hpp"

namespace mtcnvsrc {  // ::mtcnvsrc

using namespace ders;
using namespace std;

const unsigned long MaxSize=1024*1024;

struct CmdLineParser {
       int numThr;
       bool isD, isU;
       bool isE;
       sh_text masks;

       CmdLineParser(mem_pool& mp) : numThr(2), masks(nt(mp)) {}

       void parse(int argc, char** argv);
};

struct MainTask : public task {
       enum Fmt { fNo, fDos, fUnx };

       struct ErrData {
              text_buf& err;
              sh_text msg;

              ErrData(text_buf& e, sh_text m) : err(e), msg(m) {}
       };

       struct Arg {
              buf_cache bc;
              int cnt;

              Arg() : bc(MaxSize), cnt(0) {}
       };

       mem_pool& gmp;
       text_buf gerr;
       Fmt fmt;
       bool esp;
       vector<sh_wldcrd_mtchr> vmr;
       vector<int> vns;
       vector<sh_ptr<Arg> > args;

       MainTask(Fmt f, bool e, sh_text ms);

       static void addError(void* arg);
       static sh_ptr<Arg> newArg(mem_pool& mp);

       virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
       virtual void* proc_arg();
       virtual void proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers&
         to);

       void doFindFiles(mem_pool& mp, data_queue& dq, Arg& arg, const
         FindFilesMsg& msg);
       void doConvFile(mem_pool& mp, data_queue& dq, Arg& arg, const
         ConvFileMsg& msg);
       int getCnt();
};

void CmdLineParser::parse(int, char** argv)
{
 const char* usage="mtcnvsrc [-nt2] [d|u][e] mask[,mask2...]";
 mem_pool& mp=masks.pool();

 char** it=argv;
 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);

 const ch_rng ntpar("-nt");
 if (text(mp, *it).starts(ntpar)) {
    int val=atoi(*it+ntpar.size());
    if ( !(val>=1 && val<=100) ) {
       throw newExitMsgException(mp, _FLINE_,
         "-ntN must be between -nt1 and -nt100", 1);
    }

    numThr=val;

    if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 }

 bool bd=false, be=false, bu=false;
 const ch_rng deu(*it);
 for (const char* ptr=deu.beg; ptr<deu.end; ptr++)
 {
  switch (*ptr) {
         case 'd': {
              if (bd || bu) throw newExitMsgException(mp, _FLINE_, usage, 1);
              bd=true;

              break;
         }
         case 'e': {
              if (be) throw newExitMsgException(mp, _FLINE_, usage, 1);
              be=true;

              break;
         }
         case 'u': {
              if (bd || bu) throw newExitMsgException(mp, _FLINE_, usage, 1);
              bu=true;

              break;
         }
         default: throw newExitMsgException(mp, _FLINE_, usage, 1);
  }
 }

 isD=bd, isU=bu;
 isE=be;

 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 *masks=*it;

 if (*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
}

MainTask::MainTask(Fmt f, bool e, sh_text ms) :
  gmp(ms.pool()), gerr(gmp), fmt(f), esp(e)
{
 for (char* beg=ms->begin(); beg<ms->end(); ) {
     char* end=static_cast<char*>(mem_find(beg, ms->end()-beg, ','));
     if (!end) end=ms->end();

     sh_text esc_msk(nt(gmp));
     int numSep=0;

     for (const char* ptr=beg; ptr<end; ptr++) {
         if (*ptr==pathSepr) numSep++;

         if (*ptr=='\\') esc_msk->append(*ptr);
         esc_msk->append(*ptr);
     }

     vmr.push_back(new_wldcrd_mtchr(gmp, esc_msk));
     vns.push_back(numSep);

     beg=end+1;
 }
}

void MainTask::addError(void* arg)
{
 ErrData& ed=*static_cast<ErrData*>(arg);
 ed.err+'\n'+ed.msg+'\n';
}

sh_ptr<MainTask::Arg> MainTask::newArg(mem_pool& mp)
{
 mp_newbuf<Arg> buf(mp);
 return sh_ptr<Arg>(buf.pool(), buf.rls(::new(buf.get()) Arg));
}

void* MainTask::proc_arg()
{
 args.push_back(newArg(gmp));
 return args.back().get();
}

void MainTask::proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers& to)
{
 data_queue& dq=*dqv[0];
 Arg& ag=*static_cast<Arg*>(arg);

 for (;;) {
     shException exc(mp, 0);
     try {
         for (MsgIO mio(mp, dq); ; ) {
             sh_ptr<Msg> msg=mio.read();
             if (!msg.get()) break;

             switch (msg->getType()) {
                    case Msg::FindFiles: {
                         doFindFiles(mp, dq, ag, msg->to<FindFilesMsg>());
                         break;
                    }
                    case Msg::ConvFile: {
                         doConvFile(mp, dq, ag, msg->to<ConvFileMsg>());
                         break;
                    }
             }
         }

         return;
     }
     catch (shException she) { exc=she; }
     catch (...) { exc=recatchException(mp, _FLINE_); }

     ErrData ed(gerr, toTextAll(exc));
     to.invoke(addError, &ed);

     dq.set_intr(true);
     break;
 }
}

void MainTask::doFindFiles(mem_pool& mp, data_queue& dq, Arg&, const
  FindFilesMsg& msg)
{
 MsgIO mio(mp, dq);
 sh_dir shd=new_dir(mp, msg.dirName);

 for (dir::entry dent(mp); shd->find_next(dent); ) {
     if (dent.name=="." || dent.name=="..")
        continue;

     sh_text fname=shd->full_name(dent);

     if (dent.isdir) {
        FindFilesMsg m(fname);
        mio.write(m);

        continue;
     }

     assert(vmr.size()==vns.size());
     for (int i=0, end=vmr.size(); i<end; i++) {
         char* ptr=fname->end()-1;
         for (int ns=0; ptr>=fname->begin(); ptr--)
             if (*ptr==pathSepr && ++ns>vns[i]) break;

         if (vmr[i]->match(ch_rng(ptr+1, fname->end()))) {
            if (dent.size>MaxSize) {
               throw newFileErrorException(mp, _FLINE_, text_buf(mp)+
                 "Can't convert: "+dent.size+'>'+MaxSize, efbig, fname);
            }

            ConvFileMsg m(fname, dent.size);
            mio.write(m);

            break;
         }
     }
 }
}

void MainTask::doConvFile(mem_pool& mp, data_queue&, Arg& arg, const
  ConvFileMsg& msg)
{
 mem_buf buf(arg.bc.take_buf());
 int bcnt=0;

 bool drt=false;

 file fin(mp, msg.fileName, file::rdo, 0);
 buf_reader br(mp, fin, 64*1024);

 for (sh_text line(nt(mp)); br.read_line(line, false); ) {
     assert(line->size()>0);
     if (line->back()=='\n') {
        if (line->ends("\r\n")) {
           if (fmt==fUnx) {
              *(line->end()-2)='\n';
              line->uninitialized_resize(line->size()-1);

              drt=true;
           }
        }
        else {
             if (fmt==fDos) {
                line->back()='\r';
                line->append('\n');

                drt=true;
             }
        }
     }

     if (esp) {
        int ecnt=0;
        for (char* ptr=line->end()-1; ptr>=line->begin(); ptr--) {
            if (*ptr=='\n' || *ptr=='\r') ecnt++;
            else break;
        }

        int scnt=0;
        for (char* ptr=line->end()-1-ecnt; ptr>=line->begin(); ptr--) {
            if (*ptr==' ' || *ptr=='\t') scnt++;
            else break;
        }

        if (scnt) {
           memcpy(line->end()-scnt-ecnt, line->end()-ecnt, ecnt);
           line->uninitialized_resize(line->size()-scnt);

           drt=true;
        }
     }

     if (bcnt+line->size()>arg.bc.buf_size()) {
        throw newFileErrorException(mp, _FLINE_, text_buf(mp)+
          "Converted file exceeds "+arg.bc.buf_size()+" bytes", efbig,
          msg.fileName);
     }

     memcpy(static_cast<char*>(buf.get())+bcnt, line->begin(), line->size());
     bcnt+=line->size();
 }

 if (drt) {
    file fout(mp, msg.fileName, file::wro, file::crt|file::trnc);
    fout.ex_write(buf.get(), bcnt);

    arg.cnt++;
 }

 arg.bc.return_buf(buf);
}

int MainTask::getCnt()
{
 int ret=0;
 for (int i=0, end=args.size(); i<end; i++) ret+=args[i]->cnt;

 return ret;
}

}  // namespace ::mtcnvsrc

int main(int argc, char** argv)
{
 using namespace ders;
 using namespace mtcnvsrc;

 mem_pool mp;
 file err(mp, fd::err);
 file out(mp, fd::out);

 shException exc(mp, 0);
 try {
     CmdLineParser clp(mp);
     clp.parse(argc, argv);

     MainTask::Fmt fmt;
     if (clp.isD) fmt=MainTask::fDos;
     else if (clp.isU) fmt=MainTask::fUnx;
     else fmt=MainTask::fNo;

     MainTask mt(fmt, clp.isE, clp.masks);

     sh_data_queue dq=new_data_queue(mp);

     MsgIO mio(mp, *dq);
     FindFilesMsg m(nt(mp, ""));
     mio.write(m);

     sh_thread_pool tp=(clp.numThr>1) ? new_thread_pool(mp, clp.numThr-1) :
       new_thread_pool(mp);
     tp->exec(mt, dq_vec(1, dq.get()));

     out.write(text_buf(mp)+"\nProcessed: "+mt.getCnt()+'\n');

     if (mt.gerr.size()) throw newExitMsgException(mp, _FLINE_, mt.gerr, 2);

     return 0;
 }
 catch (shException she) { exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 ExitMsgException* em=exc->is<ExitMsgException>();
 if (em) {
    if (em->message->size())
       (em->exitCode ? err : out).write(text_buf(em->message)+'\n');
    return em->exitCode;
 }

 err.write(text_buf(toTextAll(exc))+'\n');
 return 2;
}

