/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Main file of mtdel program.
 */

#include <stdlib.h>
#include <ders/dir.hpp>
#include <ders/file.hpp>
#include <ders/mem.hpp>
#include <ders/path.hpp>
#include <ders/text_buf.hpp>
#include <ders/thread_pool.hpp>
#include <ders/wldcrd_mtchr.hpp>
#include "msg.hpp"

namespace mtdel {  // ::mtdel

using namespace ders;
using namespace std;

struct CmdLineParser {
       int numThr;
       sh_text masks;

       CmdLineParser(mem_pool& mp) : numThr(2), masks(nt(mp)) {}

       void parse(int argc, char** argv);
};

struct MainTask : public task {
       struct ErrData {
              text_buf& err;
              sh_text msg;

              ErrData(text_buf& e, sh_text m) : err(e), msg(m) {}
       };

       struct Stat {
              int allCnt;
              unsigned long allSize;

              Stat() : allCnt(0), allSize(0) {}
       };

       mem_pool& gmp;
       text_buf gerr;
       vector<sh_wldcrd_mtchr> vmr;
       vector<int> vns;
       list<Stat> lst;

       MainTask(sh_text ms);

       static void addError(void* arg);

       virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
       virtual void* proc_arg();
       virtual void proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers&
         to);

       void doFindFiles(mem_pool& mp, data_queue& dq, Stat& st, const
         FindFilesMsg& msg);
       void deleteFile(mem_pool& mp, Stat& st, sh_text fname, unsigned long
         size);
       Stat getStat();
};

void CmdLineParser::parse(int, char** argv)
{
 const char* usage="mtdel [-nt2] mask[,mask2...]";
 mem_pool& mp=masks.pool();

 char** it=argv;
 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);

 const ch_rng ntpar("-nt");
 if (text(mp, *it).starts(ntpar)) {
    int val=atoi(*it+ntpar.size());
    if ( !(val>=1 && val<=100) ) {
       throw newExitMsgException(mp, _FLINE_,
         "-ntN must be between -nt1 and -nt100", 1);
    }

    numThr=val;

    if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 }

 *masks=*it;

 if (*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
}

MainTask::MainTask(sh_text ms) : gmp(ms.pool()), gerr(gmp)
{
 for (char* beg=ms->begin(); beg<ms->end(); ) {
     char* end=static_cast<char*>(mem_find(beg, ms->end()-beg, ','));
     if (!end) end=ms->end();

     sh_text esc_msk(nt(gmp));
     int numSep=0;

     for (const char* ptr=beg; ptr<end; ptr++) {
         if (*ptr==pathSepr) numSep++;

         if (*ptr=='\\') esc_msk->append(*ptr);
         esc_msk->append(*ptr);
     }

     vmr.push_back(new_wldcrd_mtchr(gmp, esc_msk));
     vns.push_back(numSep);

     beg=end+1;
 }
}

void MainTask::addError(void* arg)
{
 ErrData& ed=*static_cast<ErrData*>(arg);
 ed.err+'\n'+ed.msg+'\n';
}

void* MainTask::proc_arg()
{
 lst.push_back(Stat());
 return &lst.back();
}

void MainTask::proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers& to)
{
 data_queue& dq=*dqv[0];
 Stat& st=*static_cast<Stat*>(arg);

 for (;;) {
     shException exc(mp, 0);
     try {
         for (MsgIO mio(mp, dq); ; ) {
             sh_ptr<Msg> msg=mio.read();
             if (!msg.get()) break;

             switch (msg->getType()) {
                    case Msg::FindFiles: {
                         doFindFiles(mp, dq, st, msg->to<FindFilesMsg>());
                         break;
                    }
             }
         }

         return;
     }
     catch (shException she) { exc=she; }
     catch (...) { exc=recatchException(mp, _FLINE_); }

     ErrData ed(gerr, toTextAll(exc));
     to.invoke(addError, &ed);

     dq.set_intr(true);
     break;
 }
}

void MainTask::doFindFiles(mem_pool& mp, data_queue& dq, Stat& st, const
  FindFilesMsg& msg)
{
 MsgIO mio(mp, dq);
 sh_dir shd=new_dir(mp, msg.dirName);

 for (dir::entry dent(mp); shd->find_next(dent); ) {
     if (dent.name=="." || dent.name=="..")
        continue;

     sh_text fname=shd->full_name(dent);

     if (dent.isdir) {
        FindFilesMsg m(fname);
        mio.write(m);

        continue;
     }

     assert(vmr.size()==vns.size());
     for (int i=0, end=vmr.size(); i<end; i++) {
         char* ptr=fname->end()-1;
         for (int ns=0; ptr>=fname->begin(); ptr--)
             if (*ptr==pathSepr && ++ns>vns[i]) break;

         if (vmr[i]->match(ch_rng(ptr+1, fname->end()))) {
            deleteFile(mp, st, fname, dent.size);
            break;
         }
     }
 }
}

void MainTask::deleteFile(mem_pool& mp, Stat& st, sh_text fname, unsigned long
  size)
{
 ex_remove(mp, fname);

 st.allCnt++;
 st.allSize+=size;

 file(mp, fd::out).write(text_buf(fname)+'\t'+size+'\n');
}

MainTask::Stat MainTask::getStat()
{
 Stat ret;
 for (list<Stat>::iterator it=lst.begin(), end=lst.end(); it!=end; ++it) {
     ret.allCnt+=it->allCnt;
     ret.allSize+=it->allSize;
 }

 return ret;
}

}  // namespace ::mtdel

int main(int argc, char** argv)
{
 using namespace ders;
 using namespace mtdel;

 mem_pool mp;
 file in(mp, fd::in);
 file out(mp, fd::out);
 file err(mp, fd::err);

 shException exc(mp, 0);
 try {
     CmdLineParser clp(mp);
     clp.parse(argc, argv);

     MainTask mt(clp.masks);

     out.write("Delete\n");
     for (int i=0, end=mt.vmr.size(); i<end; i++)
         out.write(text_buf(mp)+'\t'+mt.vmr[i]->mask()+'\n');
     out.write("? (Y/N) ");

     char ch;
     in.ex_read(&ch, 1);
     if ( !(ch=='y' || ch=='Y') ) return 0;

     out.write("\n");

     sh_data_queue dq=new_data_queue(mp);

     MsgIO mio(mp, *dq);
     FindFilesMsg m(nt(mp, ""));
     mio.write(m);

     sh_thread_pool tp=(clp.numThr>1) ? new_thread_pool(mp, clp.numThr-1) :
       new_thread_pool(mp);
     tp->exec(mt, dq_vec(1, dq.get()));

     MainTask::Stat st=mt.getStat();
     out.write(text_buf(mp)+"\nDeleted: "+st.allCnt+'\t'+st.allSize+'\n');

     if (mt.gerr.size()) throw newExitMsgException(mp, _FLINE_, mt.gerr, 2);

     return 0;
 }
 catch (shException she) { exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 ExitMsgException* em=exc->is<ExitMsgException>();
 if (em) {
    if (em->message->size())
       (em->exitCode ? err : out).write(text_buf(em->message)+'\n');
    return em->exitCode;
 }

 err.write(text_buf(toTextAll(exc))+'\n');
 return 2;
}

