/*
 * Copyright (C) Sergey P. Derevyago, 2008-2009.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Implements struct MainTask and related.
 */

#include "maintask.hpp"
#include <ders/dir.hpp>
#include <ders/file.hpp>
#include <ders/path.hpp>
#include "msg.hpp"

namespace {

using namespace ders;

void makeNewDirs(sh_text dir)
{
 mem_pool& mp=dir.pool();

 error err=make_dirs(mp, dir);
 if (err!=ezero && err!=eexist)
    throw newFileErrorException(mp, _FLINE_, "Can't make directory", err, dir);
}

}  // unnamed

namespace mtdirdiff {  // ::mtdirdiff

using namespace ders;

MainTask::MainTask(sh_text od, sh_text nd, sh_text dd) :
  gmp(od.pool()), gerr(gmp), argno(0), oldDir(normalize_path(gmp, od)), newDir(
  normalize_path(gmp, nd)), diffDirAdd(nt(gmp)), diffDirDel(nt(gmp)),
  diffDirMod(nt(gmp))
{
 sh_text diffDir=normalize_path(gmp, dd);
 diffDirAdd=text_buf(diffDir)+pathSepr+"add";
 diffDirDel=text_buf(diffDir)+pathSepr+"del";
 diffDirMod=text_buf(diffDir)+pathSepr+"mod";
}

void MainTask::addError(void* arg)
{
 ErrData& ed=*static_cast<ErrData*>(arg);
 ed.err+'\n'+ed.msg+'\n';
}

sh_ptr<MainTask::Stat> MainTask::newStat(mem_pool& mp)
{
 mp_newbuf<Stat> buf(mp);
 return sh_ptr<Stat>(buf.pool(), buf.rls(::new(buf.get()) Stat));
}

sh_ptr<MainTask::Arg> MainTask::newArg(mem_pool& mp)
{
 mp_newbuf<Arg> buf(mp);
 return sh_ptr<Arg>(buf.pool(), buf.rls(::new(buf.get()) Arg));
}

void* MainTask::proc_arg()
{
 assert(argno<=int(args.size()));
 if (argno==int(args.size())) args.push_back(newArg(gmp));

 return args[argno++].get();
}

void MainTask::proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers& to)
{
 data_queue& dq=*dqv[0];
 Arg& ag=*static_cast<Arg*>(arg);

 for (;;) {
     shException exc(mp, 0);
     try {
         for (MsgIO mio(mp, dq); ; ) {
             sh_ptr<Msg> msg=mio.read();
             if (!msg.get()) break;

             switch (msg->getType()) {
                    case Msg::FindFiles: {
                         doFindFiles(mp, dq, ag, msg->to<FindFilesMsg>());
                         break;
                    }
                    case Msg::Compare: {
                         doCompare(mp, dq, ag, msg->to<CompareMsg>());
                         break;
                    }
                    case Msg::CopyFile: {
                         doCopyFile(mp, dq, ag, msg->to<CopyFileMsg>());
                         break;
                    }
                    case Msg::CompareFile: {
                         doCompareFile(mp, dq, ag, msg->to<CompareFileMsg>());
                         break;
                    }
             }
         }

         return;
     }
     catch (shException she) { exc=she; }
     catch (...) { exc=recatchException(mp, _FLINE_); }

     ErrData ed(gerr, toTextAll(exc));
     to.invoke(addError, &ed);

     dq.set_intr(true);
     break;
 }
}

void MainTask::doFindFiles(mem_pool& mp, data_queue&, Arg&, const FindFilesMsg&
  msg)
{
 int pref=msg.dir->name->size();

 vector<sh_text> dirs;
 for (dirs.push_back(msg.dir->name); dirs.size(); ) {
     sh_dir shd=new_dir(mp, dirs.back());
     dirs.pop_back();

     for (dir::entry dent(mp); shd->find_next(dent); ) {
         if (dent.name=="." || dent.name=="..") continue;

         sh_text fname=shd->full_name(dent);
         if (dent.isdir) dirs.push_back(fname);

         assert(fname->size()>=pref+1);
         int beg= (fname->begin()[pref]==pathSepr) ? pref+1 : pref;
         sh_text name(nt(msg.dir->ownmp, fname->begin()+beg, fname->end()));

         if (dent.isdir) msg.dir->dirs.insert(name, 0);
         else msg.dir->files.insert(name, dent.size);
     }
 }
}

void MainTask::doCompare(mem_pool& mp, data_queue& dq, Arg&, const CompareMsg&
  msg)
{
 DirCont *dc1, *dc2;
 sh_text* dr;
 if (msg.oldToNew) {
    dc1=&oldDir;
    dc2=&newDir;
    dr=&diffDirDel;
 }
 else {
      dc1=&newDir;
      dc2=&oldDir;
      dr=&diffDirAdd;
 }

 for (int i=0; i<dc1->dirs.size(); i++) {
     const sh_text& dir=dc1->dirs.key(i);
     if (dc2->dirs.find(dir)==-1) makeNewDirs(text_buf(mp)+*dr+pathSepr+dir);
 }

 MsgIO mio(mp, dq);
 for (int i=0; i<dc1->files.size(); i++) {
     const sh_text& fil=dc1->files.key(i);
     int j=dc2->files.find(fil);

     if (j==-1) {
        CopyFileMsg m(text_buf(mp)+dc1->name+pathSepr+fil, text_buf(mp)+*dr+
          pathSepr+fil, (msg.oldToNew) ? 'd' : 'a');
        mio.write(m);
     }
     else {
          if (!msg.oldToNew) continue;

          if (dc1->files.val(i)!=dc2->files.val(j)) {
             CopyFileMsg m(text_buf(mp)+dc2->name+pathSepr+fil, text_buf(mp)+
               diffDirMod+pathSepr+fil, 'm');
             mio.write(m);
          }
          else {
               CompareFileMsg m(i);
               mio.write(m);
          }
     }
 }
}

void MainTask::doCopyFile(mem_pool& mp, data_queue&, Arg& arg, const
  CopyFileMsg& msg)
{
 file fin(mp, msg.fromName, file::rdo, 0);

 file fout(mp);
 if (!fout.open(msg.toName, file::wro, file::crt|file::trnc)) {
    make_dirs(mp, get_path(mp, msg.toName));
    fout.ex_open(msg.toName, file::wro, file::crt|file::trnc);
 }

 mem_buf buf(arg.bc.take_buf());

 for (int rd; (rd=fin.ex_read(buf.get(), arg.bc.buf_size()))>0; )
     fout.ex_write(buf.get(), rd);

 switch (msg.dest) {
        case 'a': {
             arg.st.addCnt++;
             break;
        }
        case 'd': {
             arg.st.delCnt++;
             break;
        }
        case 'm': {
             arg.st.modCnt++;
             break;
        }
        default: assert(false);
 }

 arg.bc.return_buf(buf);
}

void MainTask::doCompareFile(mem_pool& mp, data_queue& dq, Arg& arg, const
  CompareFileMsg& msg)
{
 const sh_text& name=oldDir.files.key(msg.oldPos);
 sh_text ofn=text_buf(mp)+oldDir.name+pathSepr+name, nfn=text_buf(mp)+
   newDir.name+pathSepr+name;

 file nfil(mp, nfn, file::rdo, 0), ofil(mp, ofn, file::rdo, 0);
 mem_buf nbuf(arg.bc.take_buf()), obuf(arg.bc.take_buf());

 for (;;) {
     int nrd=nfil.ex_read(nbuf.get(), arg.bc.buf_size()), ord=ofil.ex_read(
       obuf.get(), arg.bc.buf_size());
     hard_assert(nrd==ord);
     if (!nrd) break;

     if (memcmp(nbuf.get(), obuf.get(), nrd)) {
        CopyFileMsg m(nfn, text_buf(mp)+diffDirMod+pathSepr+name, 'm');
        doCopyFile(mp, dq, arg, m);

        break;
     }
 }

 arg.bc.return_buf(nbuf);
 arg.bc.return_buf(obuf);
}

sh_ptr<MainTask::Stat> MainTask::getStat(mem_pool& mp)
{
 sh_ptr<Stat> ret=newStat(mp);
 for (int i=0, end=args.size(); i<end; i++) {
     ret->addCnt+=args[i]->st.addCnt;
     ret->delCnt+=args[i]->st.delCnt;
     ret->modCnt+=args[i]->st.modCnt;
 }

 return ret;
}

}  // namespace ::mtdirdiff

