/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Main file of mtdirdiff program.
 */

#include <stdlib.h>
#include <ders/file.hpp>
#include <ders/thread_pool.hpp>
#include "maintask.hpp"

namespace mtdirdiff {  // ::mtdirdiff

using namespace ders;

struct CmdLineParser {
       int numThr;
       sh_text oldDir;
       sh_text newDir;
       sh_text diffDir;

       CmdLineParser(mem_pool& mp) : numThr(2), oldDir(nt(mp)), newDir(nt(mp)),
         diffDir(nt(mp)) {}

       void parse(int argc, char** argv);
};

void CmdLineParser::parse(int, char** argv)
{
 const char* usage="mtdirdiff [-nt2] old_dir new_dir diff_dir";
 mem_pool& mp=oldDir.pool();

 char** it=argv;
 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);

 const ch_rng ntpar("-nt");
 if (text(mp, *it).starts(ntpar)) {
    int val=atoi(*it+ntpar.size());
    if ( !(val>=1 && val<=100) ) {
       throw newExitMsgException(mp, _FLINE_,
         "-ntN must be between -nt1 and -nt100", 1);
    }

    numThr=val;

    if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 }

 *oldDir=*it;

 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 *newDir=*it;

 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 *diffDir=*it;

 if (*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
}

}  // namespace ::mtdirdiff

int main(int argc, char** argv)
{
 using namespace ders;
 using namespace mtdirdiff;

 mem_pool mp;
 file err(mp, fd::err);
 file out(mp, fd::out);

 shException exc(mp, 0);
 try {
     CmdLineParser clp(mp);
     clp.parse(argc, argv);

     sh_thread_pool tp=(clp.numThr>1) ? new_thread_pool(mp, clp.numThr-1) :
       new_thread_pool(mp);

     MainTask mt(clp.oldDir, clp.newDir, clp.diffDir);

     sh_data_queue dq=new_data_queue(mp);
     dq_vec dqv(1, dq.get());

     MsgIO mio(mp, *dq);
     {
      FindFilesMsg m1(&mt.oldDir), m2(&mt.newDir);
      mio.write(m1);
      mio.write(m2);
     }
     tp->exec(mt, dqv);

     {
      CompareMsg m1(true), m2(false);
      mio.write(m1);
      mio.write(m2);
     }
     mt.argno=0;
     tp->exec(mt, dqv);

     sh_ptr<MainTask::Stat> st=mt.getStat(mp);
     out.write(text_buf(mp)+"add\t"+st->addCnt+'\n');
     out.write(text_buf(mp)+"del\t"+st->delCnt+'\n');
     out.write(text_buf(mp)+"mod\t"+st->modCnt+'\n');

     if (mt.gerr.size()) throw newExitMsgException(mp, _FLINE_, mt.gerr, 2);

     return 0;
 }
 catch (shException she) { exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 ExitMsgException* em=exc->is<ExitMsgException>();
 if (em) {
    if (em->message->size())
       (em->exitCode ? err : out).write(text_buf(em->message)+'\n');
    return em->exitCode;
 }

 err.write(text_buf(toTextAll(exc))+'\n');
 return 2;
}

