/*
 * Copyright (C) Sergey P. Derevyago, 2008.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

/** @file
 * Main file of mtftext program.
 */

#include <vector>
#include <stdlib.h>
#include <ders/dir.hpp>
#include <ders/file.hpp>
#include <ders/text_buf.hpp>
#include <ders/thread_pool.hpp>
#include <ders/wldcrd_mtchr.hpp>
#include "msg.hpp"

namespace mtftext {  // ::mtftext

using namespace ders;

struct CmdLineParser {
       bool isS;
       int numThr;
       sh_text word;
       sh_text mask;

       CmdLineParser(mem_pool& mp) : isS(false), numThr(0), word(nt(mp)), mask(
         nt(mp)) {}

       void parse(int argc, char** argv);
};

struct MainTask : public task {
       bool exitOnErr;
       sh_text srchPatt;
       wldcrd_mtchr fileMchr;

       MainTask(bool eoe, sh_text sp, sh_text mk) : exitOnErr(eoe),
         srchPatt(sp), fileMchr(sp.pool(), mk) {}

       virtual void destroy(mem_pool& mp2) { destroy_this(this, mp2); }
       virtual void proc(mem_pool& mp, const dq_vec& dqv, void* arg, task_opers&
         to);

       void doFindFiles(mem_pool& mp, data_queue& dq, const FindFilesMsg& msg);
       void doScanFile(mem_pool& mp, data_queue& dq, const ScanFileMsg& msg);
};

void CmdLineParser::parse(int, char** argv)
{
 const char* usage="mtftext [-s] num_threads word mask";
 mem_pool& mp=word.pool();

 char** it=argv;
 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);

 if (*it==ch_rng("-s")) {
    isS=true;

    if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 }

 numThr=atoi(*it);
 if ( !(numThr>=1 && numThr<=100) ) {
    throw newExitMsgException(mp, _FLINE_, "num_threads must be in [1, 100]",
      1);
 }

 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 *word=*it;

 if (!*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
 *mask=*it;

 if (*++it) throw newExitMsgException(mp, _FLINE_, usage, 1);
}

void MainTask::proc(mem_pool& mp, const dq_vec& dqv, void*, task_opers&)
{
 data_queue& dq=*dqv[0];
 for (;;) {
     shException exc(mp, 0);
     try {
         for (MsgIO mio(mp, dq); ; ) {
             sh_ptr<Msg> msg=mio.read();
             if (!msg.get()) break;

             switch (msg->getType()) {
                    case Msg::FindFiles: {
                         doFindFiles(mp, dq, msg->to<FindFilesMsg>());
                         break;
                    }
                    case Msg::ScanFile: {
                         doScanFile(mp, dq, msg->to<ScanFileMsg>());
                         break;
                    }
             }
         }

         return;
     }
     catch (shException she) { exc=she; }
     catch (...) { exc=recatchException(mp, _FLINE_); }

     file(mp, fd::err).write(text_buf(toTextAll(exc))+'\n');

     if (exitOnErr) {
        dq.set_intr(true);
        break;
     }
 }
}

void MainTask::doFindFiles(mem_pool& mp, data_queue& dq, const FindFilesMsg&
  msg)
{
 MsgIO mio(mp, dq);
 sh_dir shd=new_dir(mp, msg.dirName);

 for (dir::entry dent(mp); shd->find_next(dent); ) {
     if (dent.name=="." || dent.name=="..")
        continue;

     sh_text fname=shd->full_name(dent);

     if (dent.isdir) {
        FindFilesMsg m(fname);
        mio.write(m);

        continue;
     }

     if (fileMchr.match(dent.name)) {
        ScanFileMsg m(fname);
        mio.write(m);
     }
 }
}

void MainTask::doScanFile(mem_pool& mp, data_queue&, const ScanFileMsg& msg)
{
 file out(mp, fd::out);

 file fin(mp, msg.fileName, file::rdo, 0);
 buf_reader br(mp, fin, 64*1024);

 sh_text line(nt(mp));
 for (int num=1; br.read_line(line); num++) {
     if (line->find(srchPatt)!=line->end())
        out.write(text_buf(mp)+msg.fileName+':'+num+':'+line+'\n');
 }
}

}  // namespace ::mtftext

int main(int argc, char** argv)
{
 using namespace ders;
 using namespace mtftext;

 mem_pool mp;
 file err(mp, fd::err);
 file out(mp, fd::out);

 shException exc(mp, 0);
 try {
     CmdLineParser clp(mp);
     clp.parse(argc, argv);

     MainTask mt(clp.isS, clp.word, clp.mask);

     sh_data_queue dq=new_data_queue(mp);

     MsgIO mio(mp, *dq);
     FindFilesMsg m(nt(mp, ""));
     mio.write(m);

     sh_thread_pool tp=(clp.numThr>1) ? new_thread_pool(mp, clp.numThr-1) :
       new_thread_pool(mp);
     tp->exec(mt, dq_vec(1, dq.get()));

     return (dq->is_intr()) ? 2 : 0;
 }
 catch (shException she) { exc=she; }
 catch (...) { exc=recatchException(mp, _FLINE_); }

 ExitMsgException* em=exc->is<ExitMsgException>();
 if (em) {
    if (em->message->size())
       (em->exitCode ? err : out).write(text_buf(em->message)+'\n');
    return em->exitCode;
 }

 err.write(text_buf(toTextAll(exc))+'\n');
 return 2;
}

